#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

import collections

dlg_bool_param = collections.namedtuple("dlg_bool_param", "description default_value")
dlg_int_param = collections.namedtuple("dlg_int_param", "description default_value")
dlg_float_param = collections.namedtuple("dlg_float_param", "description default_value")
dlg_string_param = collections.namedtuple(
    "dlg_string_param", "description default_value"
)
dlg_list_param = collections.namedtuple("dlg_list_param", "description default_value")
dlg_dict_param = collections.namedtuple("dlg_dict_param", "description default_value")


class dlg_batch_input(object):
    def __init__(self, mime_type, input_drop_class):
        self.input_drop_class = input_drop_class
        self.mime_type = mime_type


class dlg_batch_output(object):
    def __init__(self, mime_type, output_drop_class):
        self.output_drop_class = output_drop_class
        self.mime_type = mime_type


class dlg_streaming_input(object):
    def __init__(self, mime_type):
        self.mime_type = mime_type


class dlg_component(object):
    def __init__(
        self, name, description, batch_inputs, batch_outputs, streaming_inputs
    ):
        self.name = name
        self.description = description
        self.batch_inputs = batch_inputs
        self.batch_outputs = batch_outputs
        self.streaming_inputs = streaming_inputs



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import io
import os
from abc import abstractmethod

from dlg.drop import DataDROP, DEFAULT_INTERNAL_PARAMETERS
from dlg.io import MemoryIO
from dlg.meta import dlg_string_param


##
# @brief ParameterSet
# @details A set of parameters, wholly specified in EAGLE
# @par EAGLE_START
# @param category ParameterSet
# @param tag template
# @param[in] cparam/data_volume Data volume/5/Float/readwrite/False//False/Estimated size of the data contained in this node
# @param[in] cparam/group_end Group end/False/Boolean/readwrite/False//False/Is this node the end of a group?
# @param[in] cparam/mode Parset mode/"YANDA"/String/readonly/False//False/To what standard DALiuGE should filter and serialize the parameters.
# @param[in] cparam/config_data ConfigData/""/String/readwrite/False//False/Additional configuration information to be mixed in with the initial data
# @param[out] port/Config ConfigFile/File/The output configuration file
# @par EAGLE_END
class ParameterSetDROP(DataDROP):
    """
    A generic configuration file template wrapper
    This drop opens an (optional) file containing some initial configuration information, then
    appends any additional specified parameters to it, finally serving it as a data object.
    """

    config_data = b""

    mode = dlg_string_param("mode", None)

    @abstractmethod
    def serialize_parameters(self, parameters: dict, mode):
        """
        Returns a string representing a serialization of the parameters.
        """
        if mode == "YANDA":
            # TODO: Add more complex value checking
            return "\n".join(f"{x}={y}" for x, y in parameters.items())
        # Add more formats (.ini for example)
        return "\n".join(f"{x}={y}" for x, y in parameters.items())

    @abstractmethod
    def filter_parameters(self, parameters: dict, mode):
        """
        Returns a dictionary of parameters, with daliuge-internal or other parameters filtered out
        """
        if mode == "YANDA":
            forbidden_params = list(DEFAULT_INTERNAL_PARAMETERS)
            if parameters["config_data"] == "":
                forbidden_params.append("configData")
            return {
                key: val
                for key, val in parameters.items()
                if key not in DEFAULT_INTERNAL_PARAMETERS
            }
        return parameters

    def initialize(self, **kwargs):
        """
        TODO: Open input file
        """
        self.config_data = self.serialize_parameters(
            self.filter_parameters(self.parameters, self.mode), self.mode
        ).encode("utf-8")

    def getIO(self):
        return MemoryIO(io.BytesIO(self.config_data))

    @property
    def dataURL(self) -> str:
        hostname = os.uname()[1]
        return f"config://{hostname}/{os.getpid()}/{id(self.config_data)}"



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
A small, httplib-only dependent flavor of the NGAS "read" and "write" methods.
Installing the NGAS client libraries is still a not-so-trivial exercise, and
while not all DALiuGE installations have the NGAS client libraries available we
still need to access NGAS from time to time.

@author: rtobar
"""

import http.client
import logging
import urllib.request
from xml.dom.minidom import parseString

logger = logging.getLogger(__name__)


def open(
    host,
    fileId,
    port=7777,
    timeout=5,
    mode=1,
    mimeType="application/octet-stream",
    length=-1,
):
    logger.debug(
        "Opening NGAS drop %s with mode %d and length %s" % (fileId, mode, length)
    )
    if mode == 1:
        return retrieve(host, fileId, port=port, timeout=timeout)
    elif mode == 0:
        return beginArchive(
            host, fileId, port=port, timeout=timeout, mimeType=mimeType, length=length
        )
    else:
        # just return the status for that file_id
        stat = fileStatus(host, port, fileId, timeout=timeout)
        return stat


def retrieve(host, fileId, port=7777, timeout=None):
    """
    Retrieve the given fileId from the NGAS server located at `host`:`port`

    This method returns a file-like object that supports the `read` operation,
    and over which `close` must be invoked once no more data is read from it.
    """
    url = "http://%s:%d/RETRIEVE?file_id=%s" % (host, port, fileId)
    logger.debug("Issuing RETRIEVE request: %s" % (url))
    conn = urllib.request.urlopen(url)
    if conn.getcode() != http.HTTPStatus.OK:
        raise Exception(
            "Error while RETRIEVE-ing %s from %s:%d: %d %s"
            % (fileId, host, port, conn.getcode(), conn.msg)
        )
    return conn


def beginArchive(
    host, fileId, port=7777, timeout=5, length=-1, mimeType="application/octet-stream"
):
    """
    Opens a connecting to the NGAS server located at `host`:`port` and sends out
    the request for archiving the given `fileId`.

    This method returns the HTTP connection object, over which subsequential
    calls to `send` must be made with the chunks of data that need to be stored.
    Once all the data has been sent, the `finishArchive` method of this module
    should be invoked to check that all went well with the archiving.
    """
    logger.debug(
        "Issuing ARCHIVE for file %s request to: http://%s:%d" % (fileId, host, port)
    )
    conn = http.client.HTTPConnection(host, port, timeout=timeout)
    conn.putrequest("POST", "/QARCHIVE?filename=" + fileId)
    conn.putheader("Content-Type", mimeType)
    if length is not None and length >= 0:
        conn.putheader("Content-Length", length)
        # defer endheaders NGAS requires Content-Length
        conn.endheaders()
    else:
        logger.warning("Data size for %s unkown. Caching it first" % (fileId))
    return conn


def finishArchive(conn, fileId):
    """
    Checks that an archiving started by `beginArchive` went on successfully.
    """
    response = conn.getresponse()
    if response.status != http.HTTPStatus.OK:
        raise Exception(
            "Error while QARCHIVE-ing %s to %s:%d: %d %s"
            % (fileId, conn.host, conn.port, response.status, response.msg)
        )


def fileStatus(host, port, fileId, timeout=10):
    """
    Return the status as a dictionary for the file_id given
    TODO: This needs to use file_version as well, else it might
    return a value for a previous version.
    """
    url = "http://%s:%d/STATUS?file_id=%s" % (host, port, fileId)
    logger.debug("Issuing STATUS request: %s" % (url))
    try:
        conn = urllib.request.urlopen(url, timeout=timeout)
    except urllib.error.HTTPError:
        raise FileNotFoundError
    if conn.getcode() != http.HTTPStatus.OK:
        raise Exception(
            "Error while getting STATUS %s from %s:%d: %d %s"
            % (fileId, host, port, conn.getcode(), conn.msg)
        )
    dom = parseString(conn.read().decode())
    stat = dict(
        dom.getElementsByTagName("NgamsStatus")[0]
        .getElementsByTagName("DiskStatus")[0]
        .getElementsByTagName("FileStatus")[0]
        .attributes.items()
    )
    logger.debug("Returning status: %s" % (stat))
    return stat



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#

import ctypes
import functools
import logging
import multiprocessing
import queue
import threading
import six

from .. import rpc, utils
from ..ddap_protocol import AppDROPStates
from ..drop import AppDROP, BarrierAppDROP
from ..exceptions import InvalidDropException

logger = logging.getLogger(__name__)

_read_cb_type = ctypes.CFUNCTYPE(
    ctypes.c_size_t, ctypes.POINTER(ctypes.c_char), ctypes.c_size_t
)

_write_cb_type = ctypes.CFUNCTYPE(
    ctypes.c_size_t, ctypes.POINTER(ctypes.c_char), ctypes.c_size_t
)

_app_running_cb_type = ctypes.CFUNCTYPE(None)

_app_done_cb_type = ctypes.CFUNCTYPE(None, ctypes.c_int)


class CDlgInput(ctypes.Structure):
    _fields_ = [
        ("uid", ctypes.c_char_p),
        ("oid", ctypes.c_char_p),
        ("name", ctypes.c_char_p),
        ("status", ctypes.c_int),
        ("read", _read_cb_type),
    ]


class CDlgStreamingInput(ctypes.Structure):
    _fields_ = [
        ("uid", ctypes.c_char_p),
        ("oid", ctypes.c_char_p),
        ("name", ctypes.c_char_p),
    ]


class CDlgOutput(ctypes.Structure):
    _fields_ = [
        ("uid", ctypes.c_char_p),
        ("oid", ctypes.c_char_p),
        ("name", ctypes.c_char_p),
        ("write", _write_cb_type),
    ]


class CDlgApp(ctypes.Structure):
    _fields_ = [
        ("appname", ctypes.c_char_p),
        ("uid", ctypes.c_char_p),
        ("oid", ctypes.c_char_p),
        ("ranks", ctypes.POINTER(ctypes.c_int32)),
        ("n_ranks", ctypes.c_uint),
        ("inputs", ctypes.POINTER(CDlgInput)),
        ("n_inputs", ctypes.c_uint),
        ("streaming_inputs", ctypes.POINTER(CDlgStreamingInput)),
        ("n_streaming_inputs", ctypes.c_uint),
        ("outputs", ctypes.POINTER(CDlgOutput)),
        ("n_outputs", ctypes.c_uint),
        ("running", _app_running_cb_type),
        ("done", _app_done_cb_type),
        ("data", ctypes.c_void_p),
    ]

    def pack_python(self):
        out = {}
        for key, val in self._fields_:
            out[key] = repr(getattr(self, key))
        return out


def _to_c_input(i):
    """
    Convert an input drop into its corresponding C structure
    """

    input_read = i.read

    def _read(desc_, buf, n):
        x = input_read(desc_, n)
        ctypes.memmove(buf, x, len(x))
        return len(x)

    desc = i.open()
    r = _read_cb_type(functools.partial(_read, desc))
    c_input = CDlgInput(
        i.uid.encode("utf8"), i.oid.encode("utf8"), i.name.encode("utf8"), i.status, r
    )
    return desc, c_input


def _to_c_output(o):
    """
    Convert an output drop into its corresponding C structure
    """

    def _write(_o, buf, n):
        return _o.write(buf[:n])

    w = _write_cb_type(functools.partial(_write, o))
    return CDlgOutput(
        o.uid.encode("utf8"), o.oid.encode("utf8"), o.name.encode("utf8"), w
    )


def prepare_c_inputs(c_app, inputs):
    """
    Converts all inputs to its C equivalents and sets them into `c_app`
    """

    c_inputs = []
    input_closers = []
    for i in inputs:
        desc, c_input = _to_c_input(i)
        input_closers.append(functools.partial(i.close, desc))
        c_inputs.append(c_input)
    c_app.inputs = (CDlgInput * len(c_inputs))(*c_inputs)
    c_app.n_inputs = len(c_inputs)
    return input_closers


def prepare_c_outputs(c_app, outputs):
    """
    Converts all outputs to its C equivalents and sets them into `c_app`
    """

    c_outputs = [_to_c_output(o) for o in outputs]
    c_app.outputs = (CDlgOutput * len(c_outputs))(*c_outputs)
    c_app.n_outputs = len(c_outputs)


def prepare_c_ranks(c_app, ranks):
    """
    Convert the ranks list into its C equivalent and sets them to `c_app`
    """
    if ranks is None:
        # The number of ranks is
        c_app.n_ranks = 0
    else:
        c_app.ranks = (ctypes.c_int32 * len(ranks))(*ranks)
        c_app.n_ranks = len(ranks)


def run(lib, c_app, input_closers):
    """
    Invokes the `run` method on `lib` with the given `c_app`. After completion,
    all opened file descriptors are closed.
    """
    try:
        if hasattr(lib, "run2"):
            # With run2 we pass the results as a PyObject*
            run2 = lib.run2
            run2.restype = ctypes.py_object
            result = run2(ctypes.pointer(c_app))
            if isinstance(result, Exception):
                raise result
            if result:
                raise Exception(
                    "Invocation of {}:run2 returned with status {}".format(lib, result)
                )

        elif lib.run(ctypes.pointer(c_app)):
            raise Exception("Invocation of %r:run returned with status != 0" % lib)
    finally:
        for closer in input_closers:
            closer()


class InvalidLibrary(Exception):
    pass


def load_and_init(libname, oid, uid, params):
    """
    Loads and initializes `libname` with the given parameters, prepares the
    corresponding C application structure, and returns both objects
    """

    # Try with a simple name, or as full path
    from ctypes.util import find_library

    libname = find_library(libname) or libname

    lib = ctypes.cdll.LoadLibrary(libname)
    logger.info("Loaded {} as {!r}".format(libname, lib))

    one_of_functions = [["run", "run2"], ["init", "init2"]]
    for functions in one_of_functions:
        found_one = False
        for fname in functions:
            if hasattr(lib, fname):
                found_one = True
                break

        if not found_one:
            raise InvalidLibrary(
                "{} doesn't have one of the functions {}".format(libname, functions)
            )

    # Create the initial contents of the C dlg_app_info structure
    # We pass no inputs because we don't know them (and don't need them)
    # at this point yet.
    # The running and done callbacks are also NULLs
    c_app = CDlgApp(
        None,
        uid.encode("utf8"),
        oid.encode("utf8"),
        None,
        0,
        None,
        0,
        None,
        0,
        None,
        0,
        ctypes.cast(None, _app_running_cb_type),
        ctypes.cast(None, _app_done_cb_type),
        None,
    )

    if hasattr(lib, "init2"):
        # With init2 we pass the params as a PyObject*
        logger.info("Extra parameters passed to application: {}".format(params))
        init2 = lib.init2
        init2.restype = ctypes.py_object
        result = init2(ctypes.pointer(c_app), ctypes.py_object(params))
        if isinstance(result, Exception):
            raise result
        if result:
            raise InvalidLibrary(
                "{} failed during initialization (init2)".format(libname)
            )

    elif hasattr(lib, "init"):
        # Collect the rest of the parameters to pass them down to the library
        # We need to keep them in a local variable so when we expose them to
        # the app later on via pointers we still have their contents
        local_params = [
            (str(k).encode("utf8"), str(v).encode("utf8")) for k, v in params.items()
        ]
        logger.debug("Extra parameters passed to application: %r", local_params)

        # Wrap in ctypes
        str_ptr_type = ctypes.POINTER(ctypes.c_char_p)
        two_str_type = ctypes.c_char_p * 2
        app_params = [two_str_type(k, v) for k, v in local_params]
        app_params.append(None)
        params = (str_ptr_type * len(app_params))(*app_params)

        # Let the shared library initialize this app
        # If we have a list of key/value pairs that are all strings
        if lib.init(ctypes.pointer(c_app), params):
            raise InvalidLibrary(
                "{} failed during initialization (init)".format(libname)
            )

    else:
        raise InvalidLibrary(
            "{} failed during initialization. No init or init2".format(libname)
        )

    return lib, c_app


class DynlibAppBase(object):
    def initialize(self, **kwargs):
        super(DynlibAppBase, self).initialize(**kwargs)

        if "lib" not in kwargs:
            raise InvalidDropException(self, "library not specified")

        try:
            self.lib, self._c_app = load_and_init(
                kwargs.pop("lib"), self.oid, self.uid, kwargs
            )
        except InvalidLibrary as e:
            raise InvalidDropException(self, e.args[0])

        # Have we properly set the outputs in the C application structure yet?
        self._c_outputs_set = False
        self._c_outputs_setting_lock = threading.Lock()

    def _ensure_c_outputs_are_set(self):

        with self._c_outputs_setting_lock:
            if self._c_outputs_set:
                return
            prepare_c_outputs(self._c_app, self.outputs)


class DynlibStreamApp(DynlibAppBase, AppDROP):
    def initialize(self, **kwargs):
        super(DynlibStreamApp, self).initialize(**kwargs)

        # Set up callbacks for the library to signal they the application
        # is running, and that it has ended
        def _running():
            self.execStatus = AppDROPStates.RUNNING

        def _done(status):
            self.execStatus = status
            self._notifyAppIsFinished()

        self._c_app.running = _app_running_cb_type(_running)
        self._c_app.done = _app_done_cb_type(_done)

    def dataWritten(self, uid, data):
        self._ensure_c_outputs_are_set()
        app_p = ctypes.pointer(self._c_app)
        self.lib.data_written(app_p, uid.encode("utf8"), data, len(data))

    def dropCompleted(self, uid, drop_state):
        self._ensure_c_outputs_are_set()
        app_p = ctypes.pointer(self._c_app)
        self.lib.drop_completed(app_p, uid.encode("utf8"), drop_state)

    def addInput(self, inputDrop, back=True):
        super(DynlibStreamApp, self).addInput(inputDrop, back)
        self._c_app.n_inputs += 1

    def addStreamingInput(self, streamingInputDrop, back=True):
        super(DynlibStreamApp, self).addStreamingInput(streamingInputDrop, back)
        self._c_app.n_streaming_inputs += 1

    def generate_recompute_data(self):
        out = {"status": self.status}
        data = self._c_app.pack_python()
        if data is not None:
            out.update(data)
        return out


##
# @brief DynlibApp
# @details An application component run from a dynamic library
# @par EAGLE_START
# @param category DynlibApp
# @param tag template
# @param[in] cparam/libpath Library Path//String/readwrite/False//False/
#     \~English The location of the shared object/DLL that implements this application
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] cparam/group_start Group start/False/Boolean/readwrite/False//False/
#     \~English Is this node the start of a group?
# @param[in] cparam/input_error_threshold "Input error rate (%)"/0/Integer/readwrite/False//False/
#     \~English the allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param[in] cparam/n_tries Number of tries/1/Integer/readwrite/False//False/
#     \~English Specifies the number of times the 'run' method will be executed before finally giving up
# @par EAGLE_END
class DynlibApp(DynlibAppBase, BarrierAppDROP):
    """Loads a dynamic library into the current process and runs it"""

    def initialize(self, **kwargs):
        super(DynlibApp, self).initialize(**kwargs)
        self.ranks = self._getArg(kwargs, "rank", None)

    def run(self):
        input_closers = prepare_c_inputs(self._c_app, self.inputs)
        prepare_c_ranks(self._c_app, self.ranks)
        self._ensure_c_outputs_are_set()
        run(self.lib, self._c_app, input_closers)

    def generate_recompute_data(self):
        out = {"status": self.status}
        if self._c_app is None:
            return out
        else:
            out.update(self._c_app.pack_python())
            return out


class FinishSubprocess(Exception):
    pass


def _run_in_proc(*args):
    try:
        _do_run_in_proc(*args)
    except FinishSubprocess:
        pass


def _do_run_in_proc(queue, libname, oid, uid, params, inputs, outputs):
    def advance_step(f, *args, **kwargs):
        try:
            r = f(*args, **kwargs)
            queue.put(None)
            return r
        except Exception as e:
            queue.put(e)
            raise FinishSubprocess()

    # Step 1: initialise the library and return if there is an error
    lib, c_app = advance_step(load_and_init, libname, oid, uid, params)

    client = rpc.RPCClient()
    try:
        client.start()

        def setup_drop_proxies(inputs, outputs):
            to_drop_proxy = lambda x: rpc.DropProxy(client, x[0], x[1], x[2], x[3])
            inputs = [to_drop_proxy(i) for i in inputs]
            outputs = [to_drop_proxy(o) for o in outputs]
            return inputs, outputs

        inputs, outputs = advance_step(setup_drop_proxies, inputs, outputs)

        # Step 3: Finish initializing the C structure and run the application
        def do_run():
            input_closers = prepare_c_inputs(c_app, inputs)
            prepare_c_outputs(c_app, outputs)
            run(lib, c_app, input_closers)

        advance_step(do_run)
    finally:
        client.shutdown()


def get_from_subprocess(proc, q):
    """Gets elements from the queue, checking that the process is still alive"""
    while proc.is_alive():
        try:
            return q.get(timeout=0.1)
        except queue.Empty:
            pass
    raise RuntimeError("Subprocess died unexpectedly")


##
# @brief DynlibProcApp
# @details An application component run from a dynamic library in a different process
# @par EAGLE_START
# @param category DynlibProcApp
# @param tag template
# @param[in] cparam/libpath Library Path//String/readwrite/False//False/
#     \~English The location of the shared object/DLL that implements this application
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] cparam/group_start Group start/False/Boolean/readwrite/False//False/
#     \~English Is this node the start of a group?
# @param[in] cparam/input_error_threshold "Input error rate (%)"/0/Integer/readwrite/False//False/
#     \~English the allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param[in] cparam/n_tries Number of tries/1/Integer/readwrite/False//False/
#     \~English Specifies the number of times the 'run' method will be executed before finally giving up
# @par EAGLE_END
class DynlibProcApp(BarrierAppDROP):
    """Loads a dynamic library in a different process and runs it"""

    def initialize(self, **kwargs):
        super(DynlibProcApp, self).initialize(**kwargs)

        if "lib" not in kwargs:
            raise InvalidDropException(self, "library not specified")
        self.libname = kwargs.pop("lib")
        self.timeout = self._getArg(kwargs, "timeout", 600)  # 10 minutes
        self.app_params = kwargs
        self.proc = None

    def run(self):

        if not hasattr(self, "_rpc_server"):
            raise Exception("DynlibProcApp can only run within an RPC server")

        # On the sub-process we create DropProxy objects, so we need to extract
        # from our inputs/outputs their contact point (RPC-wise) information.
        # If one of our inputs/outputs is a DropProxy we already have this
        # information; otherwise we must figure it out.
        inputs = [self._get_proxy_info(i) for i in self.inputs]
        outputs = [self._get_proxy_info(o) for o in self.outputs]

        logger.info("Starting new process to run the dynlib on")
        queue = multiprocessing.Queue()
        args = (
            queue,
            self.libname,
            self.oid,
            self.uid,
            self.app_params,
            inputs,
            outputs,
        )
        self.proc = multiprocessing.Process(target=_run_in_proc, args=args)
        self.proc.start()

        try:
            steps = (
                "loading and initialising library",
                "creating DropProxy instances",
                "running the application",
            )
            for step in steps:
                logger.info("Subprocess %s", step)
                error = get_from_subprocess(self.proc, queue)
                if error is not None:
                    logger.error("Error in sub-process when " + step)
                    raise error
        finally:
            self.proc.join(self.timeout)

    def _get_proxy_info(self, x):
        if isinstance(x, rpc.DropProxy):
            return x.hostname, x.port, x.session_id, x.uid

        # TODO: we can't use the NodeManager's host directly here, as that
        #       indicates the address the different servers *bind* to
        #       (and, for example, can be 0.0.0.0)
        rpc_server = x._rpc_server
        host, port = rpc_server._rpc_host, rpc_server._rpc_port
        host = utils.to_externally_contactable_host(host, prefer_local=True)
        return (host, port, x._dlg_session.sessionId, x.uid)

    def cancel(self):
        BarrierAppDROP.cancel(self)
        try:
            self.proc.terminate()
        except:
            logger.exception("Error while terminating process %r", self.proc)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2019
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import json
import logging
import os
import time

from .. import droputils
from ..manager import constants
from ..manager.client import BaseDROPManagerClient
from ..manager.session import SessionStates
import itertools


logger = logging.getLogger(__name__)


class _StatusDumper(BaseDROPManagerClient):
    """A client that dumps the graph status each time a session is queried"""

    def __init__(self, *args, **kwargs):
        self.dump_path = kwargs.pop("dump_path")
        super(_StatusDumper, self).__init__(*args, **kwargs)

    def _dump_session_status(self, session_id):
        wgs = {
            "ssid": session_id,
            "gs": self.graph_status(session_id),
            "ts": "%.3f" % time.time(),
        }
        with open(self.dump_path, "a") as fs:
            json.dump(wgs, fs)
            fs.write(os.linesep)

    def sessions(self):
        sessions = super(_StatusDumper, self).sessions()
        for session in sessions:
            self._dump_session_status(session["sessionId"])
        return sessions

    def session(self, session_id):
        session = super(_StatusDumper, self).session(session_id)
        self._dump_session_status(session_id)
        return session


def _is_end_state(session_status):
    return session_status in (SessionStates.FINISHED, SessionStates.CANCELLED)


def _get_client(host, port, timeout, status_dump_path=None):
    kwargs = {"host": host, "port": port, "timeout": timeout}
    clazz = BaseDROPManagerClient
    if status_dump_path:
        clazz = _StatusDumper
        kwargs["dump_path"] = status_dump_path
    return clazz(**kwargs)


def _session_status(session):
    def _get(status):
        if isinstance(status, dict):
            status = [_get(s) for s in status.values()]
            if isinstance(status[0], list):
                status = list(itertools.chain(*status))
        return status

    x = _get(session["status"])
    return x


def _session_finished(session):
    session_status = _session_status(session)
    logger.debug("Session status: %r", session_status)
    if isinstance(session_status, int):
        return _is_end_state(session_status)
    return all(_is_end_state(s) for s in session_status)


def monitor_sessions(
    session_id=None,
    poll_interval=10,
    host="127.0.0.1",
    port=constants.ISLAND_DEFAULT_REST_PORT,
    timeout=30,
    status_dump_path=None,
):
    """
    Monitors the execution status of all sessions (or the single session
    specified by `session_id`) by polling `host`:`port`, and returns when they
    all have finished executing.
    """
    client = _get_client(host, port, timeout, status_dump_path)
    if session_id:
        while True:
            session = client.session(session_id)
            if _session_finished(session):
                return _session_status(session)
            time.sleep(poll_interval)
    else:
        while True:
            sessions = client.sessions()
            if all(_session_finished(s) for s in sessions):
                return {s["sessionId"]: _session_status(s) for s in sessions}
            time.sleep(poll_interval)


def monitor_sessions_repro(
    session_id=None,
    poll_interval=10,
    host="127.0.0.1",
    port=constants.ISLAND_DEFAULT_REST_PORT,
    timeout=60,
    status_dump_path=None,
):
    """
    Very similar to monitoring execution status of all (or one) session specified by `session_id`
    by polling `host`:`port`, and returns when they all have finalized their reproducibility data.
    """
    client = _get_client(host, port, timeout, status_dump_path)
    if session_id:
        while True:
            repro_status = client.session_repro_status(session_id)
            if repro_status:
                return True
            time.sleep(poll_interval)
    else:
        while True:
            sessions = client.sessions()
            if all(client.session_repro_status(s) for s in sessions):
                return {s["sessionId"]: s["repro"] for s in sessions}
            time.sleep(poll_interval)


def fetch_reproducibility(
    session_id=None,
    poll_interval=10,
    host="127.0.0.1",
    port=constants.ISLAND_DEFAULT_REST_PORT,
    timeout=60,
):
    """
    Fetches the final graph and associated reproducibility information for `session_id`.
    """
    if session_id is None:
        return {}
    client = _get_client(host, port, timeout)
    while True:
        repro_data = client.session_repro_data(session_id)
        if repro_data is not None:
            return repro_data
        time.sleep(poll_interval)


def submit(
    pg,
    host="127.0.0.1",
    port=constants.ISLAND_DEFAULT_REST_PORT,
    timeout=30,
    skip_deploy=False,
    session_id=None,
):
    """
    Submits a physical graph for execution. Additionally it waits until the
    graph has finished executing.
    """
    client = _get_client(host, port, timeout)
    session_id = session_id or "%f" % (time.time())
    completed_uids = droputils.get_roots(pg[:-1])
    with client:
        client.create_session(session_id)
        logger.info("Session %s created", session_id)
        client.append_graph(session_id, pg)
        logger.info("Graph for session %s appended", session_id)
        if not skip_deploy:
            client.deploy_session(session_id, completed_uids=completed_uids)
            logger.info("Session %s deployed", session_id)
    return session_id



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing functions to load a fully-functional DROP graph from its
full JSON representation.
"""

import collections
import importlib
import logging

from dlg.common.reproducibility.constants import ReproducibilityFlags

from numpy import isin

from . import droputils
from .apps.socket_listener import SocketListenerApp
from .common import Categories
from .ddap_protocol import DROPRel, DROPLinkType
from .drop import (
    ContainerDROP,
    InMemoryDROP,
    SharedMemoryDROP,
    FileDROP,
    NgasDROP,
    LINKTYPE_NTO1_PROPERTY,
    LINKTYPE_1TON_APPEND_METHOD,
    NullDROP,
    EndDROP,
    PlasmaDROP,
    PlasmaFlightDROP,
)
from .environmentvar_drop import EnvironmentVarDROP
from dlg.parset_drop import ParameterSetDROP
from .exceptions import InvalidGraphException
from .json_drop import JsonDROP
from .common import Categories, DropType


STORAGE_TYPES = {
    Categories.MEMORY: InMemoryDROP,
    Categories.SHMEM: SharedMemoryDROP,
    Categories.FILE: FileDROP,
    Categories.NGAS: NgasDROP,
    Categories.NULL: NullDROP,
    Categories.END: EndDROP,
    Categories.JSON: JsonDROP,
    Categories.PLASMA: PlasmaDROP,
    Categories.PLASMAFLIGHT: PlasmaFlightDROP,
    Categories.PARSET: ParameterSetDROP,
    Categories.ENVIRONMENTVARS: EnvironmentVarDROP,
}

try:
    from .s3_drop import S3DROP

    STORAGE_TYPES[Categories.S3] = S3DROP
except ImportError:
    pass

    # Dictionary for the key used to store 1-to-N relationships between DROPs
# in the the DROP specification format
__TOMANY = {
    DROPLinkType.CONSUMER: "consumers",
    DROPLinkType.STREAMING_CONSUMER: "streamingConsumers",
    DROPLinkType.INPUT: "inputs",
    DROPLinkType.STREAMING_INPUT: "streamingInputs",
    DROPLinkType.OUTPUT: "outputs",
    DROPLinkType.CHILD: "children",
    DROPLinkType.PRODUCER: "producers",
}

# Same for above, but for n-to-1 relationships
__TOONE = {DROPLinkType.PARENT: "parent"}

# Both also contain the reverse mapping
__TOMANY.update({v: k for k, v in __TOMANY.items()})
__TOONE.update({v: k for k, v in __TOONE.items()})

logger = logging.getLogger(__name__)


def addLink(linkType, lhDropSpec, rhOID, force=False):
    """
    Adds a link from `lhDropSpec` to point to `rhOID`. The link type (e.g., a
    consumer) is signaled by `linkType`.
    """

    lhOID = lhDropSpec["oid"]

    # 1-N relationship
    if linkType in __TOMANY:
        rel = __TOMANY[linkType]
        if not rel in lhDropSpec:
            relList = []
            lhDropSpec[rel] = relList
        else:
            relList = lhDropSpec[rel]
        if rhOID not in relList:
            relList.append(rhOID)
        else:
            raise Exception("DROP %s is already part of %s's %s" % (rhOID, lhOID, rel))
    # N-1 relationship, overwrite existing relationship only if `force` is specified
    elif linkType in __TOONE:
        rel = __TOONE[linkType]
        if rel and not force:
            raise Exception(
                "DROP %s already has a '%s', use 'force' to override" % (lhOID, rel)
            )
        lhDropSpec[rel] = rhOID
    else:
        raise ValueError("Cannot handle link type %d" % (linkType))

    logger.debug("Successfully linked %s and %s via '%s'", lhOID, rhOID, rel)


def removeUnmetRelationships(dropSpecList):
    unmetRelationships = []

    # Step #1: Get all OIDs
    oids = []
    for dropSpec in dropSpecList:
        oid = dropSpec["oid"]
        oid = list(oid.keys())[0] if isinstance(oid, dict) else oid
        oids.append(oid)

    # Step #2: find unmet relationships and remove them from the original
    # DROP spec, keeping track of them
    for dropSpec in dropSpecList:

        this_oid = dropSpec["oid"]
        this_oid = list(this_oid.keys())[0] if isinstance(this_oid, dict) else this_oid
        to_delete = []

        for rel in dropSpec:

            # 1-N relationships
            if rel in __TOMANY:

                link = __TOMANY[rel]

                # Find missing OIDs in this relationship and keep track of them,
                # removing them from the current DROP spec
                ds = dropSpec[rel]
                if isinstance(ds[0], dict):
                    ds = [list(d.keys())[0] for d in ds]
                missingOids = [oid for oid in ds if oid not in oids]
                for oid in missingOids:
                    unmetRelationships.append(DROPRel(oid, link, this_oid))
                    ds.remove(oid)

                # Remove the relationship list entirely if it has no elements
                if not ds:
                    to_delete.append(rel)

            # N-1 relationships
            elif rel in __TOONE:

                link = __TOONE[rel]

                # Check if OID is missing
                oid = dropSpec[rel]
                oid = list(oid.keys())[0] if isinstance(oid, dict) else oid
                if oid in oids:
                    continue

                # Keep track of missing relationship
                unmetRelationships.append(DROPRel(oid, link, this_oid))

                # Remove relationship from current DROP spec
                to_delete.append(rel)

        for rel in to_delete:
            ds = dropSpec[rel]
            ds = list(ds.keys())[0] if isinstance(ds, dict) else ds
            del ds

    return unmetRelationships


def check_dropspec(n, dropSpec):
    if "oid" not in dropSpec:
        raise InvalidGraphException(
            "Drop #%d is missing its 'oid' argument: %r" % (n, dropSpec)
        )
    if "type" not in dropSpec:
        raise InvalidGraphException(
            "Drop %s is missing its 'type' argument" % (dropSpec["oid"])
        )


def loadDropSpecs(dropSpecList):
    """
    Loads the DROP definitions from `dropSpectList`, checks that
    the DROPs are correctly specified, and return a dictionary containing
    all DROP specifications (i.e., a dictionary of dictionaries) keyed on
    the OID of each DROP. Unlike `readObjectGraph` and `readObjectGraphS`,
    this method doesn't actually create the DROPs themselves.

    Slices off graph-wise reproducibility data for later use
    """

    # Step #1: Check the DROP specs and collect them
    dropSpecs = {}
    reprodata = None
    if dropSpecList is None:
        raise InvalidGraphException("DropSpec is empty %r" % dropSpecList)
    if dropSpecList[-1].get("rmode"):
        reprodata = dropSpecList.pop()
    for n, dropSpec in enumerate(dropSpecList):

        # "type" and 'oit' are mandatory
        check_dropspec(n, dropSpec)
        dropType = dropSpec["type"]

        cf = __CREATION_FUNCTIONS[dropType]
        cf(dropSpec, dryRun=True)
        dropSpecs[dropSpec["oid"]] = dropSpec

    logger.debug("Found %d DROP definitions", len(dropSpecs))

    # Step #2: check relationships
    # TODO: shouldn't this loop be done the other way around, going through all __TOMANY
    # and __TOONE and directly address the respective dropSpec attribute?
    for dropSpec in dropSpecList:

        # 1-N relationships
        for rel in dropSpec:
            if rel in __TOMANY:

                # A KeyError will be raised if a oid has been specified in the
                # relationship list but doesn't exist in the list of DROPs
                for oid in dropSpec[rel]:
                    oid = list(oid.keys())[0] if isinstance(oid, dict) else oid
                    dropSpecs[oid]

            # N-1 relationships
            elif rel in __TOONE:
                port = (
                    list(dropSpecs[rel].keys())
                    if isinstance(dropSpecs[rel], dict)
                    else dropSpecs[rel]
                )
                # See comment above
                dropSpecs[port]

    # Done!
    return dropSpecs, reprodata


def createGraphFromDropSpecList(dropSpecList, session=None):
    logger.debug("Found %d DROP definitions", len(dropSpecList))

    # Step #1: create the actual DROPs
    drops = collections.OrderedDict()
    logger.info("Creating %d drops", len(dropSpecList))
    for n, dropSpec in enumerate(dropSpecList):

        check_dropspec(n, dropSpec)
        #        dropType = dropSpec.pop("type")
        dropType = dropSpec["type"]

        cf = __CREATION_FUNCTIONS[dropType]
        drop = cf(dropSpec, session=session)
        if session is not None:
            # Now using per-drop reproducibility setting.
            drop.reproducibility_level = ReproducibilityFlags(
                int(dropSpec.get("reprodata", {}).get("rmode", "0"))
            )
            # session.reprodata['rmode']
        drops[drop.oid] = drop

    # Step #2: establish relationships
    logger.info("Establishing relationships between drops")
    for dropSpec in dropSpecList:

        # 'oid' is mandatory
        oid = dropSpec["oid"]
        drop = drops[oid]

        for attr in dropSpec:
            # 1-N relationships
            if attr in __TOMANY:
                link = __TOMANY[attr]
                for rel in dropSpec[attr]:
                    oid = list(rel.keys())[0] if isinstance(rel, dict) else rel
                    lhDrop = drops[oid]
                    relFuncName = LINKTYPE_1TON_APPEND_METHOD[link]
                    try:
                        relFunc = getattr(drop, relFuncName)
                    except AttributeError:
                        logger.error(
                            '%r cannot be linked to %r due to missing method "%s"',
                            drop,
                            lhDrop,
                            relFuncName,
                        )
                        raise
                    relFunc(lhDrop)

            # N-1 relationships
            elif attr in __TOONE:
                link = __TOONE[attr]
                rel = dropSpec[attr]
                rel = list(rel.keys())[0] if isinstance(rel, dict) else rel
                lhDrop = drops[rel]
                propName = LINKTYPE_NTO1_PROPERTY[link]
                setattr(drop, propName, lhDrop)

    # We're done! Return the roots of the graph to the caller
    logger.info("Calculating graph roots")
    roots = []
    for drop in drops.values():
        if not droputils.getUpstreamObjects(drop):
            roots.append(drop)
    logger.info("%d graph roots found, bye-bye!", len(roots))

    return roots


def _createPlain(dropSpec, dryRun=False, session=None):
    oid, uid = _getIds(dropSpec)
    kwargs = _getKwargs(dropSpec)

    # 'storage' is mandatory
    storageType = STORAGE_TYPES[dropSpec["storage"]]
    if dryRun:
        return
    return storageType(oid, uid, dlg_session=session, **kwargs)


def _createContainer(dropSpec, dryRun=False, session=None):
    oid, uid = _getIds(dropSpec)
    kwargs = _getKwargs(dropSpec)

    # if no 'container' is specified, we default to ContainerDROP
    if DropType.CONTAINER in dropSpec:
        containerTypeName = dropSpec[DropType.CONTAINER]
        parts = containerTypeName.split(".")

        # Support old "dfms..." package names (pre-Oct2017)
        if parts[0] == "dfms":
            parts[0] = "dlg"

        module = importlib.import_module(".".join(parts[:-1]))
        containerType = getattr(module, parts[-1])
    else:
        containerType = ContainerDROP

    if dryRun:
        return

    return containerType(oid, uid, dlg_session=session, **kwargs)


def _createSocket(dropSpec, dryRun=False, session=None):
    oid, uid = _getIds(dropSpec)
    kwargs = _getKwargs(dropSpec)

    if dryRun:
        return
    return SocketListenerApp(oid, uid, dlg_session=session, **kwargs)


def _createApp(dropSpec, dryRun=False, session=None):
    oid, uid = _getIds(dropSpec)
    kwargs = _getKwargs(dropSpec)

    appName = dropSpec[DropType.APP]
    parts = appName.split(".")

    # Support old "dfms..." package names (pre-Oct2017)
    if parts[0] == "dfms":
        parts[0] = "dlg"

    try:
        module = importlib.import_module(".".join(parts[:-1]))
        appType = getattr(module, parts[-1])
    except (ImportError, AttributeError):
        raise InvalidGraphException(
            "drop %s specifies non-existent application: %s" % (oid, appName)
        )

    if dryRun:
        return
    return appType(oid, uid, dlg_session=session, **kwargs)


def _getIds(dropSpec):
    # uid is copied from oid if not explicitly given
    oid = dropSpec["oid"]
    uid = oid
    if "uid" in dropSpec:
        uid = dropSpec["uid"]
    return oid, uid


def _getKwargs(dropSpec):
    REMOVE = [
        "oid",
        "uid",
        "app",
    ]
    kwargs = dict(dropSpec)
    for kw in REMOVE:
        if kw in kwargs:
            del kwargs[kw]
    for name, spec in dropSpec.get("applicationArgs", dict()).items():
        kwargs[name] = spec["value"]
    return kwargs


__CREATION_FUNCTIONS = {
    DropType.PLAIN: _createPlain,
    DropType.CONTAINER: _createContainer,
    DropType.APP: _createApp,
    DropType.SERVICE_APP: _createApp,
    DropType.SOCKET: _createSocket,
}



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import os
import tempfile
import unittest

from dlg.drop import AbstractDROP, FileDROP
from dlg.environmentvar_drop import EnvironmentVarDROP
from dlg.utils import getDlgDir


def create_std_env_vars(name="env_vars"):
    return EnvironmentVarDROP(
        oid="a",
        uid="a",
        nm=name,
        dir_var="/HOME/",
        int_var=3,
        bool_var=False,
        float_var=0.5,
        dict_var={"first": 1, "second": "sec"},
        list_var=[1, 2.0, "3"],
    )


def create_empty_env_vars(name="env_vars"):
    return EnvironmentVarDROP(oid="b", uid="b", nm=name)


class TestEnvironmentVarDROP(unittest.TestCase):
    def test_get(self):
        """
        Tests that environment variables are read in and fetched correctly.
        """
        env_drop = create_std_env_vars()
        self.assertEqual("/HOME/", env_drop.get("dir_var"))
        self.assertEqual(3, env_drop.get("int_var"))
        self.assertEqual(False, env_drop.get("bool_var"))
        self.assertEqual(0.5, env_drop.get("float_var"))
        self.assertEqual({"first": 1, "second": "sec"}, env_drop.get("dict_var"))
        self.assertEqual([1, 2.0, "3"], env_drop.get("list_var"))
        self.assertIsNone(env_drop.get("non_var"))
        self.assertIsNone(env_drop.get("uid"))

    def test_get_empty(self):
        """
        Tests that an empty environment drop contains no environment variables.
        """
        env_drop = create_empty_env_vars()
        self.assertEqual(dict(), env_drop._variables)

    def test_get_multiple(self):
        """
        Tests the get_multiple routine for environment variables is correct
        """
        env_drop = create_std_env_vars()
        expected_vars = [
            None,
            "/HOME/",
            3,
            False,
            0.5,
            {"first": 1, "second": "sec"},
            [1, 2.0, "3"],
            None,
        ]
        query_keys = [
            "uid",
            "dir_var",
            "int_var",
            "bool_var",
            "float_var",
            "dict_var",
            "list_var",
            "non_var",
        ]
        self.assertEqual(expected_vars, env_drop.get_multiple(query_keys))

    def test_set(self):
        """
        Should currently raise un-implemented, but here for completeness
        """
        env_drop = create_std_env_vars()
        self.assertRaises(NotImplementedError, env_drop.set, "var", "val")

    def test_drop_get_single(self):
        """
        Tests the AbstractDROP fetch routine functions correctly with a single environment drop
        """
        env_drop = create_std_env_vars()
        test_drop = AbstractDROP(uid="b", oid="b")
        test_drop.addProducer(env_drop)
        self.assertEqual(
            "/HOME/", test_drop.get_environment_variable("$env_vars.dir_var")
        )
        self.assertEqual(3, test_drop.get_environment_variable("$env_vars.int_var"))
        self.assertEqual(
            False, test_drop.get_environment_variable("$env_vars.bool_var")
        )
        self.assertEqual(0.5, test_drop.get_environment_variable("$env_vars.float_var"))
        self.assertEqual(
            {"first": 1, "second": "sec"},
            test_drop.get_environment_variable("$env_vars.dict_var"),
        )
        self.assertEqual(
            [1, 2.0, "3"], test_drop.get_environment_variable("$env_vars.list_var")
        )
        self.assertEqual(
            "$env_vars.non_var", test_drop.get_environment_variable("$env_vars.non_var")
        )
        self.assertEqual(
            "$env_vars.uid", test_drop.get_environment_variable("$env_vars.uid")
        )

    def test_drop_get_multiple(self):
        """
        Tests the AbstractDROP multiple fetch routine functions correctly with a single environment
        drop
        """
        env_name = "env_vars"
        env_drop = create_std_env_vars(name=env_name)
        test_drop = AbstractDROP(uid="b", oid="b")
        test_drop.addProducer(env_drop)
        expected_vars = [
            f"${env_name}.uid",
            "/HOME/",
            3,
            False,
            0.5,
            {"first": 1, "second": "sec"},
            [1, 2.0, "3"],
            f"${env_name}.non_var",
        ]
        query_keys = [
            "uid",
            "dir_var",
            "int_var",
            "bool_var",
            "float_var",
            "dict_var",
            "list_var",
            "non_var",
        ]
        query_keys = [
            f"${env_name}.{x}" for x in query_keys
        ]  # Build queries of the correct form
        # Add some purposefully malformed vars
        extra_keys = ["dir_var", "$non_store.non_var"]
        query_keys.extend(extra_keys)
        expected_vars.extend(extra_keys)
        self.assertEqual(expected_vars, test_drop.get_environment_variables(query_keys))

    def test_drop_get_empty(self):
        """
        Tests the case where the environment drop has no name
        """
        env_name = ""
        env_drop = create_empty_env_vars(name=env_name)
        test_drop = AbstractDROP(uid="c", oid="c")
        test_drop.addProducer(env_drop)
        self.assertEqual("", test_drop.get_environment_variable(""))
        self.assertEqual("$", test_drop.get_environment_variable("$"))

    def test_drop_get_multiEnv(self):
        """
        Tests the AbstractDROP fetch routine with multiple environment drops
        """
        env1_name = "env_vars"
        env2_name = "more_vars"
        env1_drop = create_std_env_vars(name=env1_name)
        env2_drop = EnvironmentVarDROP(
            oid="d", uid="d", nm=env2_name, dir_var="/DIFFERENT/", int_var=4
        )
        test_drop = AbstractDROP(uid="c", oid="c")
        test_drop.addProducer(env1_drop)
        test_drop.addProducer(env2_drop)
        self.assertEqual(
            "/HOME/", test_drop.get_environment_variable(f"${env1_name}.dir_var")
        )
        self.assertEqual(
            "/DIFFERENT/", test_drop.get_environment_variable(f"${env2_name}.dir_var")
        )
        self.assertEqual(3, test_drop.get_environment_variable(f"${env1_name}.int_var"))
        self.assertEqual(4, test_drop.get_environment_variable(f"${env2_name}.int_var"))
        self.assertEqual(
            f"{env1_name}.int_var",
            test_drop.get_environment_variable(f"{env1_name}.int_var"),
        )
        self.assertEqual(f".int_var", test_drop.get_environment_variable(f".int_var"))
        self.assertEqual(
            f"$third_env.int_var",
            test_drop.get_environment_variable(f"$third_env.int_var"),
        )
        self.assertEqual(
            ["/HOME/", "/DIFFERENT/", 3, 4, f"${env1_name}.non_var", "$fake.var"],
            test_drop.get_environment_variables(
                [
                    f"${env1_name}.dir_var",
                    f"${env2_name}.dir_var",
                    f"${env1_name}.int_var",
                    f"${env2_name}.int_var",
                    f"${env1_name}.non_var",
                    "$fake.var",
                ]
            ),
        )

    def test_autofill_environment_vars(self):
        """
        Tests the autofilling functionality of AbstractDROP
        """
        env_drop = create_std_env_vars(name="env_vars")
        test_drop = AbstractDROP(
            oid="a",
            uid="a",
            dir_var="$env_vars.dir_var",
            int_var="$env_vars.int_var",
            non_var=set(),
        )
        test_drop.addProducer(env_drop)
        test_drop.autofill_environment_variables()
        self.assertEqual("/HOME/", test_drop.parameters["dir_var"])
        self.assertEqual(3, test_drop.parameters["int_var"])

    def test_get_dlg_vars(self):
        test_drop = AbstractDROP(
            oid="a",
            uid="a",
            dlg_root="$DLG_ROOT",
            non_dlg_var="$DLG_NONEXISTS",
            non_var=set(),
        )
        test_drop.autofill_environment_variables()
        self.assertEqual(getDlgDir(), test_drop.parameters["dlg_root"])
        self.assertEqual(getDlgDir(), test_drop.get_environment_variable("$DLG_ROOT"))
        self.assertEqual("$DLG_NONEXISTS", test_drop.parameters["non_dlg_var"])
        self.assertEqual(
            "$DLG_NONEXISTS", test_drop.get_environment_variable("$DLG_NONEXISTS")
        )

    def test_filename_integration(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            os.environ["DLG_ROOT"] = tmp_dir
            os.environ["DLG_FILE"] = "test_file"
            test_drop = FileDROP(
                oid="a", uid="a", filepath="$DLG_FILE", dirname="$DLG_ROOT"
            )
            test_drop.write(b"1234")
            self.assertEqual(tmp_dir, test_drop.dirname)
            self.assertEqual("test_file", test_drop.filepath)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2017
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""Module implementing the PyFuncApp class"""

import ast
import base64
import collections
import importlib
import inspect
import logging
import pickle

from typing import Callable
import dill
from io import StringIO
from contextlib import redirect_stdout

from dlg import droputils, utils
from dlg.drop import BarrierAppDROP
from dlg.exceptions import InvalidDropException
from dlg.meta import (
    dlg_bool_param,
    dlg_string_param,
    dlg_float_param,
    dlg_dict_param,
    dlg_component,
    dlg_batch_input,
    dlg_batch_output,
    dlg_streaming_input,
)

logger = logging.getLogger(__name__)


def serialize_data(d):
    return utils.b2s(base64.b64encode(pickle.dumps(d)))


def deserialize_data(d):
    return pickle.loads(base64.b64decode(d.encode("latin1")))


def serialize_func(f):

    if isinstance(f, str):
        parts = f.split(".")
        f = getattr(importlib.import_module(".".join(parts[:-1])), parts[-1])

    fser = dill.dumps(f)
    fdefaults = {"args": [], "kwargs": {}}
    adefaults = {"args": [], "kwargs": {}}
    a = inspect.getfullargspec(f)
    if a.defaults:
        fdefaults["kwargs"] = dict(
            zip(a.args[-len(a.defaults) :], [serialize_data(d) for d in a.defaults])
        )
        adefaults["kwargs"] = dict(
            zip(a.args[-len(a.defaults) :], [d for d in a.defaults])
        )
    logger.debug(f"Introspection of function {f}: {a}")
    logger.debug("Defaults for function %r: %r", f, adefaults)
    return fser, fdefaults


def import_using_name(app, fname):
    # If only one part check if builtin
    parts = fname.split(".")
    if len(parts) < 2:
        b = globals()['__builtins__']
        logger.debug(f"Builtins: {type(b)}")
        logger.debug(f"Function {fname}: {hasattr(b, fname)}")
        if fname in b:
            return b[fname]
        else:
            msg = "%s is not builtin and does not contain a module name" % fname
            raise InvalidDropException(app, msg)
    else:
        modname, fname = ".".join(parts[:-1]), parts[-1]
        try:
            mod = importlib.import_module(modname, __name__)
            return getattr(mod, fname)
        except ImportError as e:
            raise InvalidDropException(
                app, "Error when loading module %s: %s" % (modname, str(e))
            )
        except AttributeError:
            raise InvalidDropException(app, "Module %s has no member %s" % (modname, fname))
    


def import_using_code(code):
    return dill.loads(code)


##
# @brief PyFuncApp
# @details An application that wraps a simple python function.
# The inputs of the application are treated as the arguments of the function.
# Conversely, the output of the function is treated as the output of the
# application. If the application has more than one output, the result of
# calling the function is treated as an iterable, with each individual object
# being written to its corresponding output.
# @par EAGLE_START
# @param category PythonApp
# @param tag template
# @param[in] cparam/appclass Application Class/dlg.apps.pyfunc.PyFuncApp/String/readonly/False//False/
#     \~English Application class
# @param[in] cparam/execution_time Execution Time/5/Float/readonly/False//False/
#     \~English Estimated execution time
# @param[in] cparam/num_cpus No. of CPUs/1/Integer/readonly/False//False/
#     \~English Number of cores used
# @param[in] cparam/group_start Group start/False/Boolean/readwrite/False//False/
#     \~English Is this node the start of a group?
# @param[in] cparam/input_error_threshold "Input error rate (%)"/0/Integer/readwrite/False//False/
#     \~English The allowed failure rate of the inputs (in percent), before this component goes to ERROR state and is not executed
# @param[in] cparam/n_tries Number of tries/1/Integer/readwrite/False//False/
#     \~English Specifies the number of times the 'run' method will be executed before finally giving up
# @param[in] aparam/func_name Function Name//String/readwrite/False//False/
#     \~English Python function name
# @param[in] aparam/func_code Function Code//String/readwrite/False//False/
#     \~English Python function code, e.g. 'def function_name(args): return args'
# @param[in] aparam/pickle Pickle/false/Boolean/readwrite/False//False/
#     \~English Whether the python arguments are pickled.
# @param[in] aparam/func_defaults Function Defaults//String/readwrite/False//False/
#     \~English Mapping from argname to default value. Should match only the last part of the argnames list.
#               Values are interpreted as Python code literals and that means string values need to be quoted.
# @param[in] aparam/func_arg_mapping Function Arguments Mapping//String/readwrite/False//False/
#     \~English Mapping between argument name and input drop uids
# @par EAGLE_END
class PyFuncApp(BarrierAppDROP):
    """
    An application that wraps a simple python function.

    The inputs of the application are treated as the arguments of the function.
    Conversely, the output of the function is treated as the output of the
    application. If the application has more than one output, the result of
    calling the function is treated as an iterable, with each individual object
    being written to its corresponding output.

    Users indicate the function to be wrapped via the ``func_name`` parameter.
    In this case func_name needs to specify a funtion in the standard form

    ``module.function``

    and the module needs to be accessible on the PYTHONPATH of the DALiuGE
    engine. Note that the engine is expanding the standard PYTHONPATH with
    DLG_ROOT/code. That directory is always available, even if the engine is
    running in a docker container.

    Otherwise, users can also *send* over the python code using the ``func_code``
    parameter. The code needs to be base64-encoded and produced with the marshal
    module of the same Python version used to run DALiuGE.

    Both inputs and outputs are (de-)serialized using the pickle protocol if the value
    of the respective boolean component parameter is set to True. This is also
    applied to func_defaults and func_arg_mappings.

    In addition to the input mapping the implementation also allows to set defaults
    both in the function itself and in a logical graph. If set in the logical graph
    using the func_defaults parameter, the defaults need to be specified as a
    dictionary of the form

    ``{"kwargs":{"kw1_name":kw1_value, "kw2_name":kw2_value}, "args":[arg1, arg2]}``

    The positional onlyargs will be used in order of appearance.
    """

    component_meta = dlg_component(
        "PyFuncApp",
        "Py Func App.",
        [dlg_batch_input("binary/*", [])],
        [dlg_batch_output("binary/*", [])],
        [dlg_streaming_input("binary/*")],
    )

    func_name = dlg_string_param("func_name", None)

    # func_code = dlg_bytes_param("func_code", None) # bytes or base64 string

    pickle = dlg_bool_param("pickle", True)

    func_arg_mapping = dlg_dict_param("func_arg_mapping", {})

    func_defaults = dlg_dict_param("func_defaults", {})

    f: Callable
    fdefaults: dict

    def _init_func_defaults(self):
        """
        Inititalize self.func_defaults dictionary from values provided.
        Multiple options exist and some are here for compatibility.
        """
        logger.debug(f"Starting evaluation of func_defaults: {self.func_defaults}")
        if (
            isinstance(self.func_defaults, dict)
            and len(self.func_defaults) > 0
            and list(self.func_defaults.keys()) == ["kwargs", "args"]
        ):
            # we bring everything back to just kwargs, because positional args are messy
            # NOTE: This means that positional ONLY arguments won't work, but those are not used
            # too often.
            for arg in self.func_defaults["args"]:
                self.func_defaults["kwargs"][arg] = arg
                self.func_defaults = self.func_defaults["kwargs"]
        elif (
            isinstance(self.func_defaults, dict)
            and "kwargs" in self.func_defaults
            and isinstance(self.func_defaults["kwargs"], dict)
        ):
            self.func_defaults = self.func_defaults["kwargs"]
        # we came all this way, now assume that any resulting dict is correct
        if not isinstance(self.func_defaults, dict):
            logger.error(
                f"Wrong format or type for function defaults for "
                + "{self.f.__name__}: {self.func_defaults}, {type(self.func_defaults)}"
            )
            raise ValueError
        if self.pickle:
            # only values are pickled, get them unpickled
            for name, value in self.func_defaults.items():
                self.func_defaults[name] = deserialize_data(value)

        # set the function defaults from introspection
        if self.arguments:
            self.fn_npos = len(self.arguments.args) - self.fn_ndef
            self.fn_defaults = {
                name: None for name in self.arguments.args[: self.fn_npos]
            }
            logger.debug(f"initialized fn_defaults with {self.fn_defaults}")
            # deal with args and kwargs
            kwargs = (
                dict(zip(self.arguments.args[self.fn_npos :], self.arguments.defaults))
                if self.arguments.defaults
                else {}
            )
            self.fn_defaults.update(kwargs)
            logger.debug(f"fn_defaults updated with {kwargs}")
            # deal with kwonlyargs
            if self.arguments.kwonlydefaults:
                kwonlyargs = dict(
                    zip(self.arguments.kwonlyargs, self.arguments.kwonlydefaults)
                )
                self.fn_defaults.update(kwonlyargs)
                logger.debug(f"fn_defaults updated with {kwonlyargs}")

            self.fn_posargs = self.arguments.args[
                : self.fn_npos
            ]  # positional arg names

    def initialize(self, **kwargs):
        """
        The initialization of a function component is mainly dealing with mapping
        inputs and provided applicationArgs to the function arguments. All of this
        should be driven by matching names, but currently that is not being done.
        """
        BarrierAppDROP.initialize(self, **kwargs)

        self._applicationArgs = self._getArg(kwargs, "applicationArgs", {})

        self.func_code = self._getArg(kwargs, "func_code", None)

        # check for function definition arguments in applicationArgs
        self.func_def_keywords = [
            "func_code",
            "func_name",
            "func_arg_mapping",
            "pickle",
            "func_defaults"
            ]
        for kw in self.func_def_keywords:
            dum_arg = new_arg = "gIbbERiSH:askldhgol"
            if kw in self._applicationArgs:  # these are the preferred ones now
                if isinstance(
                    self._applicationArgs[kw]["value"], bool
                ):  # always transfer booleans
                    new_arg = self._applicationArgs.pop(kw)
                elif (
                    self._applicationArgs[kw]["value"]
                    or self._applicationArgs[kw]["precious"]
                ):
                    # only transfer if there is a value or precious is True
                    new_arg = self._applicationArgs.pop(kw)

            if new_arg != dum_arg:
                logger.debug(f"Setting {kw} to {new_arg['value']}")
                # we allow python expressions as values, means that strings need to be quoted
                self.__setattr__(kw, new_arg["value"])

        self.num_args = len(
            self._applicationArgs
        )  # number of additional arguments provided

        if not self.func_name and not self.func_code:
            raise InvalidDropException(
                self, "No function specified (either via name or code)"
            )

        # Lookup function or import bytecode as a function
        if not self.func_code:
            self.f = import_using_name(self, self.func_name)
        else:
            if not isinstance(self.func_code, bytes):
                self.func_code = base64.b64decode(self.func_code.encode("utf8"))
            self.f = import_using_code(self.func_code)
        # make sure defaults are dicts
        if isinstance(self.func_defaults, str):
            self.func_defaults = ast.literal_eval(self.func_defaults)
        if isinstance(self.func_arg_mapping, str):
            self.func_arg_mapping = ast.literal_eval(self.func_arg_mapping)

        self.arguments = inspect.getfullargspec(self.f)
        logger.debug(f"Function inspection revealed {self.arguments}")
        # we don't want to mess with the 'self' argument
        if self.arguments.args.count('self'): 
            self.arguments.args.remove('self')
        self.fn_nargs = len(self.arguments.args)
        self.fn_ndef = len(self.arguments.defaults) if self.arguments.defaults else 0
        self._init_func_defaults()
        logger.info(f"Args summary for '{self.func_name}':")
        logger.info(f"Args: {self.arguments.args}")
        logger.info(f"Args defaults:  {self.arguments.defaults}")
        logger.info(f"Args positional: {self.arguments.args[:self.fn_npos]}")
        logger.info(f"Args keyword: {self.arguments.args[self.fn_npos:]}")
        logger.info(f"Args supplied:  {self.func_defaults}")
        logger.info(f"VarArgs allowed:  {self.arguments.varargs}")
        logger.info(f"VarKwds allowed:  {self.arguments.varkw}")

        # Mapping between argument name and input drop uids
        logger.debug(f"Input mapping: {self.func_arg_mapping}")
        self._recompute_data = {}

    def run(self):
        """
        Function positional and keyword argument treatment:

        Function arguments can be provided in four different ways:
        1) Through an input port
        2) By specifying ApplicationArgs (one for each argument)
        3) By specifying a func_defaults dictionary in the ComponentParameters
        4) Through defaults at the time of function definition

        The priority follows the list above with input ports overruling the others.
        Function arguments in Python can be passed as positional, kw-value, positional
        only, kw-value only, and catch-all args and kwargs, which don't provide any
        hint about the names of accepted parameters. All of them are now supported. If
        positional arguments or kw-value arguments are provided by the user, but are 
        not explicitely defined in the function signiture AND args and/or kwargs are
        allowed then these arguments are passed to the function. For args this is
        somewhat risky, since the order is relevant and in this code derived from the
        order defined in the graph (same order as defined in the component description).

        Input ports will NOT be used by order (anymore), but by the IdText (name field
        in EAGLE) of the port. Since each input port requires an associated data drop,
        this provides a unique mapping. This also allows to pass values to any function
        argument through a port.

        Function argument values as well as the function code can be provided in
        serialised (pickle) form by setting the 'pickle' flag. Note that this flag
        is valid for all arguments and the code (if specified) in a global way.
        """

        # Inputs are un-pickled and treated as the arguments of the function
        # Their order must be preserved, so we use an OrderedDict
        if self.pickle:
            all_contents = lambda x: pickle.loads(x)
        else:
            all_contents = lambda x: ast.literal_eval(
                droputils.allDropContents(x).decode("utf-8")
            )

        inputs = collections.OrderedDict()
        for uid, drop in self._inputs.items():
            contents = droputils.allDropContents(drop)
            # allow for Null DROPs to be passed around
            inputs[uid] = all_contents(contents) if contents else None

        self.funcargs = {}

        # Keyword arguments are made up of the default values plus the inputs
        # that match one of the keyword argument names
        # if defaults dict has not been specified at all we'll go ahead anyway
        n_args = len(self.func_defaults)
        argnames = self.arguments.args

        # use explicit mapping of inputs to arguments first
        # TODO: Required by dlg_delayed?? Else, we should really not do this.
        kwargs = {
            name: inputs.pop(uid)
            for name, uid in self.func_arg_mapping.items()
            if name in self.func_defaults or name not in argnames
        }
        logger.debug(f"updating funcargs with {kwargs}")
        self.funcargs = kwargs

        # Fill arguments with rest of inputs
        logger.debug(f"available inputs: {inputs}")

        # if we have named ports use the inputs with
        # the correct UIDs
        logger.debug(f"Parameters found: {self.parameters}")
        posargs = self.arguments.args[:self.fn_npos]
        kwargs = {}
        self.pargs = []
        pargsDict = collections.OrderedDict(zip(posargs,[None]*len(posargs)))  # Initialize pargs dictionary
        if ('inputs' in self.parameters and isinstance(self.parameters['inputs'][0], dict)):
            logger.debug(f"Using named ports to identify inputs: "+\
                    f"{self.parameters['inputs']}")
            for i in range(min(len(inputs),self.fn_nargs +\
                len(self.arguments.kwonlyargs))):
                # key for final dict is value in named ports dict
                key = list(self.parameters['inputs'][i].values())[0]
                # value for final dict is value in inputs dict
                value = inputs[list(self.parameters['inputs'][i].keys())[0]]
                if key in posargs:
                    pargsDict.update({key:value})
                else:
                    kwargs.update({key:value})
        else:
            for i in range(min(len(inputs),self.fn_nargs)):
                kwargs.update({self.arguments.args[i]: list(inputs.values())[i]})

        logger.debug(f"updating funcargs with input ports {kwargs}")
        self.funcargs.update(kwargs)

        # Try to get values for still missing positional arguments from Application Args
        if "applicationArgs" in self.parameters:
            appArgs = self.parameters["applicationArgs"]  # we'll pop them
            _dum = [appArgs.pop(k) for k in self.func_def_keywords if k in appArgs]
            for pa in posargs:
                if pa not in self.funcargs:
                    if pa in appArgs and pa != 'self':
                        arg = appArgs.pop(pa)
                        value = arg['value']
                        ptype = arg['type']
                        if ptype in ["Complex", "Json"]:
                            try:
                                value = ast.literal_eval(value)
                            except:
                                pass
                        pargsDict.update({
                            pa:
                            value
                        })
                    elif pa != 'self' and pa not in pargsDict:
                        logger.warning(f"Required positional argument '{pa}' not found!")
            logger.debug(f"updating posargs with {list(pargsDict.values())}")
            self.pargs.extend(list(pargsDict.values()))

            # Try to get values for still missing kwargs arguments from Application kws
            kwargs = {}
            kws = self.arguments.args[self.fn_npos:]
            for ka in kws:
                if ka not in self.funcargs:
                    if ka in appArgs:
                        arg = appArgs.pop(ka)
                        value = arg['value']
                        ptype = arg['type']
                        if ptype in ["Complex", "Json"]:
                            try:
                                value = ast.literal_eval(value)
                            except:
                                pass
                        kwargs.update({
                            ka:
                            value
                        })
                    else:
                        logger.warning(f"Keyword argument '{ka}' not found!")
            logger.debug(f"updating funcargs with {kwargs}")
            self.funcargs.update(kwargs)
            vparg = []
            vkarg = {}
            logger.debug(f"Remaining AppArguments {appArgs}")
            for arg in appArgs:
                if appArgs[arg]['type'] in ['Json', 'Complex']:
                    value = ast.literal_eval(appArgs[arg]['value'])
                else:
                    value = appArgs[arg]['value']
                if appArgs[arg]['positional']:
                    vparg.append(value)
                else:
                    vkarg.update({arg:value})

            # any remaining application arguments will be used for vargs and vkwargs
            if self.arguments.varargs:
                self.pargs.extend(vparg)
            if self.arguments.varkw:
                self.funcargs.update(vkarg)

        # Fill rest with default arguments if there are any more
        kwargs = {}
        for kw in self.func_defaults.keys():
            value = self.func_defaults[kw]
            if kw not in self.funcargs:
                kwargs.update({kw: value})
        logger.debug(f"updating funcargs with {kwargs}")
        self.funcargs.update(kwargs)
        self._recompute_data["args"] = self.funcargs.copy()
        logger.debug(f"Running {self.func_name} with *{self.pargs} **{self.funcargs}")

        # we capture and log whatever is produced on STDOUT
        capture = StringIO()
        with redirect_stdout(capture):
            result = self.f(*self.pargs, **self.funcargs)
        logger.info(f"Captured output from function app '{self.func_name}': {capture.getvalue()}")
        logger.debug(f"Finished execution of {self.func_name}.")

        # Depending on how many outputs we have we treat our result
        # as an iterable or as a single object. Each result is pickled
        # and written to its corresponding output
        self.write_results(result)

    def write_results(self, result):
        outputs = self.outputs
        if len(outputs) > 0:
            if len(outputs) == 1:
                result = [result]
            for r, o in zip(result, outputs):
                p = pickle.dumps(r)
                if self.pickle:
                    logger.debug(f"Writing pickeled result {type(r)} to {o}")
                    o.write(pickle.dumps(r))  # @UndefinedVariable
                else:
                    o.write(repr(r).encode("utf-8"))

    def generate_recompute_data(self):
        return self._recompute_data



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import unittest
from dlg.parset_drop import ParameterSetDROP
from dlg.droputils import allDropContents


class test_ParameterSetDROP(unittest.TestCase):
    kwargs = {
        "mode": None,
        "Cimager": 2,
        "StringParam": "param",
        "Boolparam": True,
        "config_data": "",
        "iid": -1,
        "rank": 0,
        "consumers": ["drop-1", "drop-2"],
    }

    def test_initialize(self):
        yanda_kwargs = dict(self.kwargs)
        yanda_kwargs["mode"] = "YANDA"

        yanda_parset = ParameterSetDROP(oid="a", uid="a", **yanda_kwargs)
        standard_parset = ParameterSetDROP(oid="b", uid="b", **self.kwargs)

        yanda_output = "Cimager=2\nStringParam=param\nBoolparam=True"
        standard_output = (
            "mode=None\n"
            + yanda_output
            + "\nconfig_data=\niid=-1\nrank=0\nconsumers=['drop-1', 'drop-2']"
        )

        yanda_parset.setCompleted()
        standard_parset.setCompleted()

        self.assertEqual(yanda_output, allDropContents(yanda_parset).decode("utf-8"))
        self.assertEqual(
            standard_output, allDropContents(standard_parset).decode("utf-8")
        )



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
from abc import abstractmethod, ABCMeta
from http.client import HTTPConnection
from multiprocessing.sharedctypes import Value
from overrides import overrides
import io
import logging
import os
import sys
import urllib.parse
from abc import abstractmethod, ABCMeta
from typing import Optional, Union

from . import ngaslite
from .apps.plasmaflight import PlasmaFlightClient

import pyarrow
import pyarrow.plasma as plasma

if sys.version_info >= (3, 8):
    from dlg.shared_memory import DlgSharedMemory


logger = logging.getLogger(__name__)


class OpenMode:
    """
    Open Mode for Data Drops
    """

    OPEN_WRITE, OPEN_READ = range(2)


class DataIO:
    """
    A class used to read/write data stored in a particular kind of storage in an
    abstract way. This base class simply declares a number of methods that
    deriving classes must actually implement to handle different storage
    mechanisms (e.g., a local filesystem or an NGAS server).

    An instance of this class represents a particular piece of data. Thus at
    construction time users must specify a storage-specific unique identifier
    for the data that this object handles (e.g., a filename in the case of a
    DataIO class that works with local filesystem storage, or a host:port/fileId
    combination in the case of a class that works with an NGAS server).

    Once an instance has been created it can be opened via its `open` method
    indicating an open mode. If opened with `OpenMode.OPEN_READ`, only read
    operations will be allowed on the instance, and if opened with
    `OpenMode.OPEN_WRITE` only writing operations will be allowed.
    """

    __metaclass__ = ABCMeta
    _mode: Optional[OpenMode]

    def __init__(self):
        self._mode = None

    def open(self, mode: OpenMode, **kwargs):
        """
        Opens the underlying storage where the data represented by this instance
        is stored. Depending on the value of `mode` subsequent calls to
        `self.read` or `self.write` will succeed or fail.
        """
        self._mode = mode
        self._desc = self._open(**kwargs)

    def write(self, data, **kwargs) -> int:
        """
        Writes `data` into the storage
        """
        if self._mode is None:
            raise ValueError("Writing operation attempted on closed DataIO object")
        if self._mode == OpenMode.OPEN_READ:
            raise ValueError("Writing operation attempted on write-only DataIO object")
        return self._write(data, **kwargs)

    def read(self, count: int, **kwargs):
        """
        Reads `count` bytes from the underlying storage.
        """
        if self._mode is None:
            raise ValueError("Reading operation attempted on closed DataIO object")
        if self._mode == OpenMode.OPEN_WRITE:
            raise ValueError("Reading operation attempted on write-only DataIO object")
        return self._read(count, **kwargs)

    def close(self, **kwargs):
        """
        Closes the underlying storage where the data represented by this
        instance is stored, freeing underlying resources.
        """
        if self._mode is None:
            return
        self._close()
        self._mode = None

    def size(self, **kwargs) -> int:
        """
        Returns the current total size of the underlying stored object. If the
        storage class does not support this it is supposed to return -1.
        """
        return self._size(**kwargs)

    def isOpened(self):
        """
        Returns true if the io is currently opened for read or write.
        """
        return self._mode is not None

    @abstractmethod
    def exists(self) -> bool:
        """
        Returns `True` if the data represented by this DataIO exists indeed in
        the underlying storage mechanism
        """

    @abstractmethod
    def delete(self):
        """
        Deletes the data represented by this DataIO
        """

    def buffer(self) -> Union[memoryview, bytes, bytearray, pyarrow.Buffer]:
        """
        Gets a buffer protocol compatible object of the drop data.
        This may be a zero-copy view of the data or a copy depending
        on whether the drop stores data in cpu memory or not.
        """

    @abstractmethod
    def _open(self, **kwargs):
        pass

    @abstractmethod
    def _read(self, count, **kwargs):
        pass

    @abstractmethod
    def _write(self, data, **kwargs) -> int:
        pass

    @abstractmethod
    def _close(self, **kwargs):
        pass

    @abstractmethod
    def _size(self, **kwargs) -> int:
        pass


class NullIO(DataIO):
    """
    A DataIO that stores no data
    """

    def _open(self, **kwargs):
        return None

    def _read(self, count=4096, **kwargs):
        return None

    def _write(self, data, **kwargs) -> int:
        return len(data)

    def _close(self, **kwargs):
        pass

    @overrides
    def _size(self, **kwargs) -> int:
        """
        Size is always 0 for this storage class
        """
        return 0

    @overrides
    def exists(self) -> bool:
        return True

    @overrides
    def delete(self):
        pass


class ErrorIO(DataIO):
    """
    An DataIO method that throws exceptions if any of its methods is invoked
    """

    @overrides
    def _open(self, **kwargs):
        raise NotImplementedError()

    @overrides
    def _read(self, count=4096, **kwargs):
        raise NotImplementedError()

    @overrides
    def _write(self, data, **kwargs) -> int:
        raise NotImplementedError()

    @overrides
    def _close(self, **kwargs):
        raise NotImplementedError()

    @overrides
    def _size(self, **kwargs) -> int:
        raise NotImplementedError()

    @overrides
    def exists(self) -> bool:
        raise NotImplementedError()

    @overrides
    def delete(self):
        raise NotImplementedError()


class MemoryIO(DataIO):
    """
    A DataIO class that reads/write from/into the BytesIO object given at
    construction time
    """

    _desc: io.BytesIO  # TODO: This might actually be a problem

    def __init__(self, buf: io.BytesIO, **kwargs):
        super().__init__()
        self._buf = buf

    def _open(self, **kwargs):
        if self._mode == OpenMode.OPEN_WRITE:
            return self._buf
        elif self._mode == OpenMode.OPEN_READ:
            # TODO: potentially wasteful copy
            return io.BytesIO(self._buf.getbuffer())
        else:
            raise ValueError()

    @overrides
    def _write(self, data, **kwargs) -> int:
        self._desc.write(data)
        return len(data)

    @overrides
    def _read(self, count=4096, **kwargs):
        return self._desc.read(count)

    @overrides
    def _close(self, **kwargs):
        if self._mode == OpenMode.OPEN_READ:
            self._desc.close()
        # If we're writing we don't close the descriptor because it's our
        # self._buf, which won't be readable afterwards

    @overrides
    def _size(self, **kwargs) -> int:
        """
        Return actual size of user data rather than the whole Python object
        """
        return self._buf.getbuffer().nbytes

    @overrides
    def exists(self) -> bool:
        return not self._buf.closed

    @overrides
    def delete(self):
        self._buf.close()

    @overrides
    def buffer(self) -> memoryview:
        # TODO: This may also be an issue
        return self._buf.getbuffer()


class SharedMemoryIO(DataIO):
    """
    A DataIO class that writes to a shared memory buffer
    """

    def __init__(self, uid, session_id, **kwargs):
        super().__init__()
        self._name = f"{session_id}_{uid}"
        self._written = 0
        self._pos = 0
        self._buf = None

    @overrides
    def _open(self, **kwargs):
        self._pos = 0
        if self._buf is None:
            if self._mode == OpenMode.OPEN_WRITE:
                self._written = 0
            self._buf = DlgSharedMemory(self._name)
        return self._buf

    @overrides
    def _write(self, data, **kwargs) -> int:
        total_size = len(data) + self._written
        if total_size > self._buf.size:
            self._buf.resize(total_size)
            self._buf.buf[self._written : total_size] = data
            self._written = total_size
        else:
            self._buf.buf[self._written : total_size] = data
            self._written = total_size
            self._buf.resize(total_size)
            # It may be inefficient to resize many times, but assuming data is written 'once' this is
            # might be tolerable and guarantees that the size of the underlying buffer is tight.
        return len(data)

    @overrides
    def _read(self, count=4096, **kwargs):
        if self._pos == self._buf.size:
            return None
        start = self._pos
        end = self._pos + count
        end = min(end, self._buf.size)
        out = self._buf.buf[start:end]
        self._pos = end
        return out

    @overrides
    def _close(self, **kwargs):
        if self._mode == OpenMode.OPEN_WRITE:
            self._buf.resize(self._written)
        self._buf.close()
        self._buf = None

    @overrides
    def _size(self, **kwargs) -> int:
        return self._buf.size

    @overrides
    def exists(self) -> bool:
        return self._buf is not None

    @overrides
    def delete(self):
        self._close()


class FileIO(DataIO):
    """
    A file-based implementation of DataIO
    """

    _desc: io.BufferedReader

    def __init__(self, filename, **kwargs):
        super().__init__()
        self._fnm = filename

    def _open(self, **kwargs) -> io.BufferedReader:
        flag = "r" if self._mode is OpenMode.OPEN_READ else "w"
        flag += "b"
        return open(self._fnm, flag)

    @overrides
    def _read(self, count=4096, **kwargs):
        return self._desc.read(count)

    @overrides
    def _write(self, data, **kwargs) -> int:
        self._desc.write(data)
        return len(data)

    @overrides
    def _close(self, **kwargs):
        self._desc.close()

    @overrides
    def _size(self, **kwargs) -> int:
        return os.path.getsize(self._fnm)

    def getFileName(self):
        """
        Returns the drop filename
        """
        return self._fnm

    @overrides
    def exists(self) -> bool:
        return os.path.isfile(self._fnm)

    @overrides
    def delete(self):
        os.unlink(self._fnm)

    @overrides
    def buffer(self) -> bytes:
        return self._desc.read(-1)


class NgasIO(DataIO):
    """
    A DROP whose data is finally stored into NGAS. Since NGAS doesn't
    support appending data to existing files, we store all the data temporarily
    in a file on the local filesystem and then move it to the NGAS destination
    """

    def __init__(
        self,
        hostname,
        fileId,
        port=7777,
        ngasConnectTimeout=2,
        ngasTimeout=2,
        length=-1,
        mimeType="application/octet-stream",
    ):

        # Check that we actually have the NGAMS client libraries
        try:
            from ngamsPClient import ngamsPClient  # @UnusedImport @UnresolvedImport
        except:
            logger.error("No NGAMS client libs found, cannot use NgasIO")
            raise

        super(NgasIO, self).__init__()
        self._ngasSrv = hostname
        self._ngasPort = port
        self._ngasConnectTimeout = ngasConnectTimeout
        self._ngasTimeout = ngasTimeout
        self._fileId = fileId
        self._length = length
        self._mimeType = mimeType

    def _getClient(self):
        from ngamsPClient import ngamsPClient  # @UnresolvedImport

        return ngamsPClient.ngamsPClient(
            self._ngasSrv, self._ngasPort, self._ngasTimeout
        )

    @overrides
    def _open(self, **kwargs):
        if self._mode == OpenMode.OPEN_WRITE:
            # The NGAS client API doesn't have a way to continually feed an ARCHIVE
            # request with data. Thus the only way we can currently archive data
            # into NGAS is by accumulating it all on our side and finally
            # sending it over.
            self._buf = b""
            self._writtenDataSize = 0
        return self._getClient()

    @overrides
    def _close(self, **kwargs):
        client = self._desc
        if self._mode == OpenMode.OPEN_WRITE:
            reply, msg, _, _ = client._httpPost(
                client.getHost(),
                client.getPort(),
                "QARCHIVE",
                self._mimeType,
                dataRef=self._buf,
                pars=[["filename", self._fileId]],
                dataSource="BUFFER",
                dataSize=self._writtenDataSize,
            )
            self._buf = None
            if reply.http_status != 200:
                # Probably msg is not enough, we need to unpack the status XML doc
                # from the returning data and extract the real error message from
                # there
                raise Exception(reply.message)

        # Release the reference to _desc so the client object gets destroyed
        del self._desc

    @overrides
    def _read(self, count, **kwargs):
        # Read data from NGAS and give it back to our reader
        self._desc.retrieve2File(self._fileId, cmd="QRETRIEVE")

    @overrides
    def _write(self, data, **kwargs) -> int:
        if type(data) == bytes:
            self._buf += str(data)
        else:
            self._buf += data
        self._writtenDataSize += len(data)
        return len(data)

    @overrides
    def exists(self) -> bool:
        import ngamsLib  # @UnresolvedImport

        status = self._getClient().sendCmd("STATUS", pars=[["fileId", self._fileId]])
        return status.getStatus() == ngamsLib.ngamsCore.NGAMS_SUCCESS

    def fileStatus(self):
        import ngamsLib  # @UnresolvedImport

        # status = self._getClient().sendCmd('STATUS', pars=[['fileId', self._fileId]])
        status = self._getClient.fileStatus("STATUS?file_id=%s" % self._fileId)
        if status.getStatus() != ngamsLib.ngamsCore.NGAMS_SUCCESS:
            raise FileNotFoundError
        fs = dict(
            status.getDiskStatusList()[0]
            .getFileObjList()[0]
            .genXml()
            .attributes.items()
        )
        return fs

    @overrides
    def _size(self, **kwargs) -> int:
        return self._writtenDataSize

    def delete(self):
        pass  # We never delete stuff from NGAS


class NgasLiteIO(DataIO):
    """
    An IO class whose data is finally stored into NGAS. It uses the ngaslite
    module of DALiuGE instead of the full client-side libraries provided by NGAS
    itself, since they might not be installed everywhere.

    The `ngaslite` module doesn't support the STATUS command yet, and because of
    that this class will throw an error if its `exists` method is invoked.
    """

    _desc: HTTPConnection

    def __init__(
        self,
        hostname,
        fileId,
        port=7777,
        ngasConnectTimeout=2,
        ngasTimeout=2,
        length=-1,
        mimeType="application/octet-stream",
    ):
        super(NgasLiteIO, self).__init__()
        self._ngasSrv = hostname
        self._ngasPort = port
        self._ngasConnectTimeout = ngasConnectTimeout
        self._ngasTimeout = ngasTimeout
        self._fileId = fileId
        self._length = length
        self._mimeType = mimeType

    def _is_length_unknown(self):
        return self._length is None or self._length < 0

    def _getClient(self):
        return ngaslite.open(
            self._ngasSrv,
            self._fileId,
            port=self._ngasPort,
            timeout=self._ngasTimeout,
            mode=self._mode,
            mimeType=self._mimeType,
            length=self._length,
        )

    def _open(self, **kwargs):
        if self._mode == OpenMode.OPEN_WRITE:
            if self._is_length_unknown():
                # NGAS does not support HTTP chunked writes and thus it requires the Content-length
                # of the whole fileObject to be known and sent up-front as part of the header. Thus
                # is size is not provided all data will be buffered IN MEMORY and only sent to NGAS
                # when finishArchive is called.
                self._buf = b""
                self._writtenDataSize = 0
        return self._getClient()

    def _close(self, **kwargs):
        if self._mode == OpenMode.OPEN_WRITE:
            conn = self._desc
            if self._is_length_unknown():
                # If length wasn't known up-front we first send Content-Length and then the buffer here.
                conn.putheader("Content-Length", len(self._buf))
                conn.endheaders()
                logger.debug("Sending data for file %s to NGAS" % (self._fileId))
                conn.send(self._buf)
                self._buf = None
            else:
                logger.debug(
                    f"Length is known, assuming data has been sent ({self._fileId}, {self._length})"
                )
            ngaslite.finishArchive(conn, self._fileId)
            conn.close()
        else:
            response = self._desc
            response.close()

    def _read(self, count=4096, **kwargs):
        return self._desc.read(count)

    def _write(self, data, **kwargs) -> int:
        if self._is_length_unknown():
            self._buf += data
        else:
            self._desc.send(data)
        logger.debug("Wrote %s bytes" % len(data))
        return len(data)

    def exists(self) -> bool:
        raise NotImplementedError("This method is not supported by this class")

    def fileStatus(self):
        logger.debug("Getting status of file %s", self._fileId)
        return ngaslite.fileStatus(self._ngasSrv, self._ngasPort, self._fileId)

    @overrides
    def delete(self):
        pass  # We never delete stuff from NGAS


def IOForURL(url):
    """
    Returns a DataIO instance that handles the given URL for reading. If no
    suitable DataIO class can be found to handle the URL, `None` is returned.
    """
    url = urllib.parse.urlparse(url)
    io = None
    if url.scheme == "file":
        hostname = url.netloc
        filename = url.path
        if (
            hostname == "localhost"
            or hostname == "127.0.0.1"
            or hostname == os.uname()[1]
        ):
            io = FileIO(filename)
    elif url.scheme == "null":
        io = NullIO()
    elif url.scheme == "ngas":
        networkLocation = url.netloc
        if ":" in networkLocation:
            hostname, port = networkLocation.split(":")
            port = int(port)
        else:
            hostname = networkLocation
            port = 7777
        fileId = url.path[1:]  # remove the trailing slash
        try:
            io = NgasIO(hostname, fileId, port)
        except:
            logger.warning("NgasIO not available, using NgasLiteIO instead")
            io = NgasLiteIO(hostname, fileId, port)

    logger.debug("I/O chosen for dataURL %s: %r", url, io)

    return io


class PlasmaIO(DataIO):
    """
    A shared-memory IO reader/writer implemented using plasma store
    memory buffers. Note: not compatible with PlasmaClient put()/get()
    which performs data pickling before writing.
    """

    _desc: plasma.PlasmaClient

    def __init__(
        self,
        object_id: plasma.ObjectID,
        plasma_path="/tmp/plasma",
        expected_size: Optional[int] = None,
        use_staging=False,
    ):
        """Initializer
        Args:
            object_id (plasma.ObjectID): 20 bytes unique object id
            plasma_path (str, optional): The socket file path visible to all shared processes. Defaults to "/tmp/plasma".
            expected_size (Optional[int], optional) Total size of data to allocate to buffer if known. Defaults to None.
            use_staging (bool, optional): Whether to stream first to a resizable staging buffer. Defaults to False.
        """
        super().__init__()
        self._plasma_path = plasma_path
        self._object_id = object_id
        self._reader = None
        self._writer = None
        # treat sizes <1 as None
        self._expected_size = (
            expected_size if expected_size and expected_size > 0 else None
        )
        self._buffer_size = 0
        self._use_staging = use_staging

    @overrides
    def _open(self, **kwargs):
        return plasma.connect(self._plasma_path)

    @overrides
    def _close(self, **kwargs):
        if self._writer:
            if self._use_staging:
                self._desc.put_raw_buffer(self._writer.getbuffer(), self._object_id)
                self._writer.close()
            else:
                self._desc.seal(self._object_id)
                self._writer.close()
        if self._reader:
            self._reader.close()

    def _read(self, count, **kwargs):
        if not self._reader:
            [data] = self._desc.get_buffers([self._object_id])
            self._reader = pyarrow.BufferReader(data)
        return self._reader.read1(count)

    @overrides
    def _write(self, data, **kwargs) -> int:
        """
        Writes data into the PlasmaIO reserved buffer.
        If use_staging is False and expected_size is None, only a single write is allowed.
        If use_staging is False and expected_size is > 0, multiple writes up to expected_size is allowed.
        If use_staging is True, any number of writes may occur with a small performance penalty.
        """
        # NOTE: data must be a collection of bytes for len to represent the buffer bytesize
        assert isinstance(
            data, Union[memoryview, bytes, bytearray, pyarrow.Buffer].__args__
        )
        databytes = data.nbytes if isinstance(data, memoryview) else len(data)

        if self._use_staging:
            if not self._writer:
                # write into a resizable staging buffer
                self._writer = io.BytesIO()
        else:
            if not self._writer:
                # write directly into fixed size plasma buffer
                self._buffer_size = (
                    self._expected_size
                    if self._expected_size is not None
                    else databytes
                )
                plasma_buffer = self._desc.create(self._object_id, self._buffer_size)
                self._writer = pyarrow.FixedSizeBufferWriter(plasma_buffer)
            if self._writer.tell() + databytes > self._buffer_size:
                raise IOError(
                    "".join(
                        [
                            f"attempted to write {self._writer.tell() + databytes} ",
                            f"bytes to plasma buffer of size {self._buffer_size}, ",
                            "consider using staging or expected_size argument",
                        ]
                    )
                )

        self._writer.write(data)
        return len(data)

    @overrides
    def _size(self, **kwargs) -> int:
        return self._buffer_size

    @overrides
    def exists(self) -> bool:
        return self._object_id in self._desc.list()

    @overrides
    def delete(self):
        self._desc.delete([self._object_id])

    @overrides
    def buffer(self) -> memoryview:
        [data] = self._desc.get_buffers([self._object_id])
        return memoryview(data)


class PlasmaFlightIO(DataIO):
    """
    A plasma drop managed by an arrow flight network protocol
    """

    _desc: PlasmaFlightClient

    def __init__(
        self,
        object_id: plasma.ObjectID,
        plasma_path="/tmp/plasma",
        flight_path: Optional[str] = None,
        expected_size: Optional[int] = None,
        use_staging=False,
    ):
        super().__init__()
        self._object_id = object_id
        self._plasma_path = plasma_path
        self._flight_path = flight_path
        self._reader = None
        self._writer = None
        # treat sizes <1 as None
        self._expected_size = (
            expected_size if expected_size and expected_size > 0 else None
        )
        self._buffer_size = 0
        self._use_staging = use_staging

    def _open(self, **kwargs):
        return PlasmaFlightClient(socket=self._plasma_path)

    def _close(self, **kwargs):
        if self._writer:
            if self._use_staging:
                self._desc.put_raw_buffer(self._writer.getbuffer(), self._object_id)
                self._writer.close()
            else:
                if self._expected_size != self._writer.tell():
                    logger.debug(
                        f"written {self._writer.tell()} but expected {self._expected_size} bytes"
                    )
                self._desc.seal(self._object_id)
        if self._reader:
            self._reader.close()

    def _read(self, count, **kwargs):
        if not self._reader:
            data = self._desc.get_buffer(self._object_id, self._flight_path)
            self._reader = pyarrow.BufferReader(data)
        return self._reader.read1(count)

    def _write(self, data, **kwargs) -> int:

        # NOTE: data must be a collection of bytes for len to represent the buffer bytesize
        assert isinstance(
            data, Union[memoryview, bytes, bytearray, pyarrow.Buffer].__args__
        )
        databytes = data.nbytes if isinstance(data, memoryview) else len(data)
        if not self._writer:
            if self._use_staging:
                # stream into resizeable buffer
                logger.warning(
                    "Using dynamically sized Plasma buffer. Performance may be reduced."
                )
                self._writer = io.BytesIO()
            else:
                # write directly to fixed size plasma buffer
                self._buffer_size = (
                    self._expected_size
                    if self._expected_size is not None
                    else databytes
                )
                plasma_buffer = self._desc.create(self._object_id, self._buffer_size)
                self._writer = pyarrow.FixedSizeBufferWriter(plasma_buffer)
        self._writer.write(data)
        return len(data)

    @overrides
    def exists(self) -> bool:
        return self._desc.exists(self._object_id, self._flight_path)

    @overrides
    def _size(self, **kwargs) -> int:
        return self._buffer_size

    @overrides
    def delete(self):
        pass

    @overrides
    def buffer(self) -> memoryview:
        return self._desc.get_buffer(self._object_id, self._flight_path)



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2014
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing the NodeManager, which directly manages DROP instances, and
thus represents the bottom of the DROP management hierarchy.
"""

import abc
import collections
import logging
from psutil import cpu_count
import multiprocessing.pool
import os
import queue
import sys
import threading
import time

from glob import glob
from . import constants
from .drop_manager import DROPManager
from .session import Session

if sys.version_info >= (3, 8):
    from .shared_memory_manager import DlgSharedMemoryManager
from .. import rpc, utils
from ..ddap_protocol import DROPStates
from ..drop import AppDROP, InMemoryDROP, SharedMemoryDROP
from ..exceptions import (
    NoSessionException,
    SessionAlreadyExistsException,
    DaliugeException,
)
from ..lifecycle.dlm import DataLifecycleManager

logger = logging.getLogger(__name__)


class NMDropEventListener(object):
    def __init__(self, nm, session_id):
        self._nm = nm
        self._session_id = session_id

    def handleEvent(self, event):
        event.session_id = self._session_id
        self._nm.publish_event(event)


class LogEvtListener(object):
    def handleEvent(self, event):
        if event.type == "status":
            logger.debug(
                "Drop uid=%s, oid=%s changed to state %s",
                event.uid,
                event.oid,
                event.status,
            )
        elif event.type == "execStatus":
            logger.debug(
                "AppDrop uid=%s, oid=%s changed to execState %s",
                event.uid,
                event.oid,
                event.execStatus,
            )


class ErrorStatusListener(object):
    """An event listener that passes down the erroneous drop to an error handler"""

    def __init__(self, session, error_listener):
        self._session = session
        self._error_listener = error_listener

    def handleEvent(self, evt):
        if evt.type == "status" and evt.status == DROPStates.ERROR:
            self._error_listener.on_error(self._session.drops[evt.uid])


def _load(obj, callable_attr):
    """
    Returns object (a python object or a string denoting a class within a
    python module only if it has the indicated attribute and it is callable.
    """
    if isinstance(obj, str):
        obj = utils.get_symbol(obj)()
    if not hasattr(obj, callable_attr) or not callable(getattr(obj, callable_attr)):
        raise ValueError(
            "%r doesn't contain an %s attribute that can be called"
            % (obj, callable_attr)
        )
    return obj


class NodeManagerBase(DROPManager):
    """
    Base class for a DROPManager that creates and holds references to DROPs.

    A NodeManagerBase is the ultimate responsible of handling DROPs. It does so not
    directly, but via Sessions, which represent and encapsulate separate,
    independent DROP graph executions. All DROPs created by the
    different Sessions are also given to a common DataLifecycleManager, which
    takes care of expiring them when needed and replicating them.

    Since a NodeManagerBase can handle more than one session, in principle only one
    NodeManagerBase is needed for each computing node, thus its name.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(
        self,
        useDLM=False,
        dlgPath=None,
        error_listener=None,
        event_listeners=[],
        max_threads=0,
        logdir=utils.getDlgLogsDir(),
    ):

        self._dlm = DataLifecycleManager() if useDLM else None
        self._sessions = {}
        self.logdir = logdir

        # dlgPath may contain code added by the user with possible
        # DROP applications
        if dlgPath:
            dlgPath = os.path.expanduser(dlgPath)
            if os.path.isdir(dlgPath):
                logger.info("Adding %s to the system path", dlgPath)
                sys.path.append(dlgPath)
                # we also add underlying site-packages dir to support
                # the --prefix installation of code
                pyVer = f"{sys.version_info.major}.{sys.version_info.minor}"
                extraPath = f"{dlgPath}/lib/python{pyVer}/site-packages"
                logger.info("Adding %s to the system path", extraPath)
                sys.path.append(extraPath)

        # Error listener used by users to deal with errors coming from specific
        # Drops in whatever way they want. This is a specific case of an event
        # listener, so we add it together with the rest of the user-supplied
        # event listeners
        self._error_listener = (
            _load(error_listener, "on_error") if error_listener else None
        )
        self._event_listeners = [_load(l, "handleEvent") for l in event_listeners]

        # Start thread pool
        self._threadpool = None
        if max_threads == -1:  # default use all CPU cores
            max_threads = cpu_count(logical=False)
        else:  # never more than 200
            max_threads = max(min(max_threads, 200), 1)
        if sys.version_info >= (3, 8):
            self._memoryManager = DlgSharedMemoryManager()
            if max_threads > 1:
                logger.info("Initializing thread pool with %d threads", max_threads)
                self._threadpool = multiprocessing.pool.ThreadPool(
                    processes=max_threads
                )

        # Event handler that only logs status changes
        debugging = logger.isEnabledFor(logging.DEBUG)
        self._logging_event_listener = LogEvtListener() if debugging else None

        # Start the mix-ins
        self.start()

    @abc.abstractmethod
    def start(self):
        """
        Starts any background task required by this Node Manager
        """

    @abc.abstractmethod
    def shutdown(self):
        """
        Stops any pending background task run by this Node Manager
        """

    @abc.abstractmethod
    def subscribe(self, host, port):
        """
        Subscribes this Node Manager to events published in from ``host``:``port``
        """

    @abc.abstractmethod
    def publish_event(self, evt):
        """
        Publishes the event ``evt`` for other Node Managers to receive it
        """

    @abc.abstractmethod
    def get_rpc_client(self, hostname, port):
        """
        Creates an RPC client connected to the node manager running in
        ``host``:``port``, and its closing method, as a 2-tuple.
        """

    def deliver_event(self, evt):
        """
        Method called by subclasses when a new event has arrived through the
        subscription mechanism.
        """
        if not evt.session_id in self._sessions:
            logger.warning(
                "No session %s found, event (%s) will be dropped"
                % (evt.session_id, evt.type)
            )
            return
        self._sessions[evt.session_id].deliver_event(evt)

    def _check_session_id(self, session_id):
        if session_id not in self._sessions:
            raise NoSessionException(session_id)

    def createSession(self, sessionId):
        if sessionId in self._sessions:
            raise SessionAlreadyExistsException(sessionId)
        self._sessions[sessionId] = Session(sessionId, nm=self)
        logger.info("Created session %s", sessionId)

    def getSessionStatus(self, sessionId):
        self._check_session_id(sessionId)
        return self._sessions[sessionId].status

    def getSessionReproStatus(self, sessionId):
        return self._sessions[sessionId].reprostatus

    def getGraphReproData(self, sessionId):
        return self._sessions[sessionId].reprodata

    def linkGraphParts(self, sessionId, lhOID, rhOID, linkType):
        self._check_session_id(sessionId)
        self._sessions[sessionId].linkGraphParts(lhOID, rhOID, linkType)

    def addGraphSpec(self, sessionId, graphSpec):
        self._check_session_id(sessionId)
        self._sessions[sessionId].addGraphSpec(graphSpec)

    def getGraphStatus(self, sessionId):
        self._check_session_id(sessionId)
        return self._sessions[sessionId].getGraphStatus()

    def getGraph(self, sessionId):
        self._check_session_id(sessionId)
        #  TODO: Ensure returns reproducibility data.
        return self._sessions[sessionId].getGraph()

    def getLogDir(self):
        return self.logdir

    def deploySession(self, sessionId, completedDrops=[]):
        self._check_session_id(sessionId)
        session = self._sessions[sessionId]
        if hasattr(self, "_memoryManager"):
            self._memoryManager.register_session(sessionId)

        def foreach(drop):
            drop.autofill_environment_variables()
            if self._threadpool is not None:
                drop._tp = self._threadpool
                if isinstance(drop, InMemoryDROP):
                    drop._sessID = sessionId
                    self._memoryManager.register_drop(drop.uid, sessionId)
            elif isinstance(drop, SharedMemoryDROP):
                if sys.version_info < (3, 8):
                    raise NotImplementedError(
                        "Shared memory is not implemented when using Python < 3.8"
                    )
                drop._sessID = sessionId
                self._memoryManager.register_drop(drop.uid, sessionId)
            if self._dlm:
                self._dlm.addDrop(drop)

            # Remote event forwarding
            evt_listener = NMDropEventListener(self, sessionId)
            if isinstance(drop, AppDROP):
                drop.subscribe(evt_listener, "producerFinished")
            else:
                drop.subscribe(evt_listener, "dropCompleted")

            # Purely for logging purposes
            log_evt_listener = self._logging_event_listener
            if log_evt_listener:
                drop.subscribe(log_evt_listener, "status")
                drop.subscribe(log_evt_listener, "reproducibility")
                if isinstance(drop, AppDROP):
                    drop.subscribe(log_evt_listener, "execStatus")

        # Add user-supplied listeners
        listeners = self._event_listeners[:]
        if self._error_listener:
            listeners.append(ErrorStatusListener(session, self._error_listener))

        session.deploy(
            completedDrops=completedDrops, event_listeners=listeners, foreach=foreach
        )

    def cancelSession(self, sessionId):
        logger.info("Cancelling session: %s", sessionId)
        self._check_session_id(sessionId)
        self._sessions[sessionId].cancel()

    def destroySession(self, sessionId):
        logger.info("Destroying session: %s", sessionId)
        self._check_session_id(sessionId)
        session = self._sessions.pop(sessionId)
        if hasattr(self, "_memoryManager"):
            self._memoryManager.shutdown_session(sessionId)
        session.destroy()

    def getSessionIds(self):
        return list(self._sessions.keys())

    def getGraphSize(self, sessionId):
        self._check_session_id(sessionId)
        session = self._sessions[sessionId]
        return len(session._graph)

    def trigger_drops(self, sessionId, uids):
        self._check_session_id(sessionId)
        t = threading.Thread(
            target=self._sessions[sessionId].trigger_drops,
            name="Drop trigger",
            args=(uids,),
        )
        t.start()

    def add_node_subscriptions(self, sessionId, relationships):

        logger.debug("Received subscription information: %r", relationships)
        self._check_session_id(sessionId)
        self._sessions[sessionId].add_node_subscriptions(relationships)

        # Set up event channels subscriptions
        for nodesub in relationships:

            host = nodesub
            events_port = constants.NODE_DEFAULT_EVENTS_PORT
            if type(nodesub) is tuple:
                host, events_port, _ = nodesub

            # TODO: we also have to unsubscribe from them at some point
            self.subscribe(host, events_port)

    def has_method(self, sessionId, uid, mname):
        self._check_session_id(sessionId)
        return self._sessions[sessionId].has_method(uid, mname)

    def get_drop_property(self, sessionId, uuid, prop_name):
        self._check_session_id(sessionId)
        return self._sessions[sessionId].get_drop_property(uuid, prop_name)

    def call_drop(self, sessionId, uid, method, *args):
        self._check_session_id(sessionId)
        return self._sessions[sessionId].call_drop(uid, method, *args)

    def shutdown(self):
        if hasattr(self, "_threadpool") and self._threadpool is not None:
            self._threadpool.close()
            self._threadpool.join()


class ZMQPubSubMixIn(object):
    """
    ZeroMQ-based event publisher and subscriber.

    Event publishing and event reception are done in their own separate
    threads, where the externally-facing ZeroMQ sockets are created and used.

    Events to be published are fed into the publishing thread via a safe-thread
    Queue object (self._events_out), enabling any local thread to publish events
    without having to worry about ZeroMQ thread-safeness.

    The event reception thread not only *receives* events, but also updates the
    subscription socket to connect to new peers. These updates are fed via a
    Queue object (self._subscriptions), enabling any local thread to indicate a
    new peer to subscribe to in a thread-safe manner.

    Note that we investigated not using Queue objects to communicate between
    threads, and use inproc:// ZeroMQ sockets instead. This works, but at a
    cost: all threads putting values into these sockets would need to check,
    each time they use a socket in any manner, if the Context object is still
    valid and hasn't been closed (or alternatively if self._pubsub_running is
    still True). Our experience with this alternative was not satisfactory, and
    therefore we went for a Queue-based thread communication model, making the
    handling of ZeroMQ resources simpler.
    """

    subscription = collections.namedtuple("subscription", "endpoint finished_evt")

    def __init__(self, host, events_port):
        self._events_host = host
        self._events_port = events_port

    def start(self):
        self._pubsub_running = True
        super(ZMQPubSubMixIn, self).start()
        self._events_in = queue.Queue()
        self._events_out = queue.Queue()
        self._subscriptions = queue.Queue()

        # Starts background threads, but wait until their sockets are created
        timeout = 30
        self._event_publisher = self._start_thread(
            self._publish_events, "Evt pub", timeout
        )
        self._event_receiver = self._start_thread(
            self._receive_events, "Evt recv", timeout
        )
        self._event_deliverer = self._start_thread(self._deliver_events, "Evt delivery")

    def _start_thread(self, target, name, timeout=None):
        evt = threading.Event() if timeout else None
        args = (evt,) if evt else ()
        t = threading.Thread(target=target, name=name, args=args)
        t.start()
        if evt and not evt.wait(timeout):
            raise Exception("Failed to start %s thread in %d seconds" % (name, timeout))
        return t

    def shutdown(self):
        super(ZMQPubSubMixIn, self).shutdown()
        self._pubsub_running = False
        self._event_deliverer.join()
        self._event_publisher.join()
        self._event_receiver.join()
        logger.info("ZeroMQ event publisher/subscriber finished")

    def publish_event(self, evt):
        self._events_out.put(evt)

    def subscribe(self, host, port):
        timeout = 5
        finished_evt = threading.Event()
        endpoint = "tcp://%s:%d" % (utils.zmq_safe(host), port)
        self._subscriptions.put(ZMQPubSubMixIn.subscription(endpoint, finished_evt))
        if not finished_evt.wait(timeout):
            raise DaliugeException(
                "ZMQ subscription not achieved within %d seconds" % (timeout,)
            )
        logger.info("Subscribed for events originating from %s", endpoint)

    def _publish_events(self, sock_created):
        import zmq

        pub = self._context.socket(zmq.PUB)  # @UndefinedVariable
        pub.set_hwm(0)  # Never drop messages that should be sent
        endpoint = "tcp://%s:%d" % (
            utils.zmq_safe(self._events_host),
            self._events_port,
        )
        pub.bind(endpoint)
        logger.info("Publishing events via ZeroMQ on %s", endpoint)
        sock_created.set()

        while self._pubsub_running:

            try:
                obj = self._events_out.get_nowait()
            except queue.Empty:
                time.sleep(0.01)
                continue

            while self._pubsub_running:
                try:
                    pub.send_pyobj(obj, flags=zmq.NOBLOCK)  # @UndefinedVariable
                    break
                except zmq.error.Again:
                    logger.debug("Got an 'Again' when publishing event")
                    time.sleep(0.01)
                    continue
        pub.close()

    def _deliver_events(self):
        while self._pubsub_running:
            try:
                evt = self._events_in.get_nowait()
                self.deliver_event(evt)
            except queue.Empty:
                time.sleep(0.01)

    def _receive_events(self, sock_created):
        import zmq
        from zmq.utils.monitor import recv_monitor_message

        sub = self._context.socket(zmq.SUB)  # @UndefinedVariable
        sub_endpoints = set()
        sub.setsockopt(zmq.SUBSCRIBE, b"")  # @UndefinedVariable
        sub_monitor = sub.get_monitor_socket()
        sock_created.set()

        pending_connections = {}
        while self._pubsub_running:

            # A new subscription has been requested
            try:
                subscription = self._subscriptions.get_nowait()
                if subscription.endpoint in sub_endpoints:
                    subscription.finished_evt.set()
                else:
                    sub.connect(subscription.endpoint)
                    pending_connections[
                        subscription.endpoint
                    ] = subscription.finished_evt
            except queue.Empty:
                pass

            try:
                msg = recv_monitor_message(sub_monitor, flags=zmq.NOBLOCK)
                if msg["event"] != zmq.EVENT_CONNECTED:
                    continue
                endpoint = utils.b2s(msg["endpoint"])
                sub_endpoints.add(endpoint)
                finished_evt = pending_connections.pop(endpoint)
                finished_evt.set()
            except zmq.error.Again:
                pass

            try:
                evt = sub.recv_pyobj(flags=zmq.NOBLOCK)  # @UndefinedVariable
                self._events_in.put(evt)
            except zmq.error.Again:
                time.sleep(0.01)
            except Exception:
                # Figure out what to do here
                logger.exception(
                    "Something bad happened in %s:%d to ZMQ :'(",
                    self._events_host,
                    self._events_port,
                )
                break

        # Flush pending connection events to avoid callers hanging out forever
        for evt in pending_connections:
            evt.set()

        sub_monitor.close()
        sub.close()


# So far we currently support ZMQ only for event publishing
EventMixIn = ZMQPubSubMixIn


# Load the corresponding RPC classes and finish the construciton of NodeManager
class RpcMixIn(rpc.RPCClient, rpc.RPCServer):
    pass


# Final NodeManager class
class NodeManager(EventMixIn, RpcMixIn, NodeManagerBase):
    def __init__(
        self,
        useDLM=True,
        dlgPath=utils.getDlgPath(),
        error_listener=None,
        event_listeners=[],
        max_threads=0,
        logdir=utils.getDlgLogsDir(),
        host=None,
        rpc_port=constants.NODE_DEFAULT_RPC_PORT,
        events_port=constants.NODE_DEFAULT_EVENTS_PORT,
    ):
        # We "just know" that our RpcMixIn will have a create_context static
        # method, which in reality means we are using the ZeroRPCServer class
        self._context = RpcMixIn.create_context()
        host = host or "127.0.0.1"
        EventMixIn.__init__(self, host, events_port)
        RpcMixIn.__init__(self, host, rpc_port)
        NodeManagerBase.__init__(
            self, useDLM, dlgPath, error_listener, event_listeners, max_threads, logdir
        )

    def shutdown(self):
        super(NodeManager, self).shutdown()
        self._context.term()



#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Module containing the DALiuGE Daemon class and command-line entry point to use it
"""

import functools
import json
import logging
import signal
import socket
import sys
import threading

import bottle
import zeroconf as zc

from . import constants, client
from .. import utils
from ..restserver import RestServer

logger = logging.getLogger(__name__)


def get_tool():
    # This import is performed at runtime to avoid a circular dependency
    # at import time with the tool module, which imports this module
    # to make it available as a 'dlg' command
    from ..common import tool

    return tool


def _get_address(zeroconf_service_info):
    if tuple(map(int, zc.__version__.split(".")))[:2] >= (0, 23):
        return zeroconf_service_info.addresses[0]
    return zeroconf_service_info.address


class DlgDaemon(RestServer):
    """
    The DALiuGE Daemon

    The DALiuGE Daemon is the long-running process that we assume is always
    available for contacting, and that acts as the bootstrapping of the whole
    system. It exposes a REST API through which users can start the different
    Drop Managers (node, dataisland and master) and query their status.
    Optionally it can also start automatically the node manager (default: yes)
    and the master manager (default: no) at creation time.
    """

    def __init__(self, master=False, noNM=False, disable_zeroconf=False, verbosity=0):

        super(DlgDaemon, self).__init__()

        self._shutting_down = False
        self._verbosity = verbosity

        # The three processes we run
        self._nm_proc = None
        self._dim_proc = None
        self._mm_proc = None

        # Zeroconf for NM and MM
        self._zeroconf = None if disable_zeroconf else zc.Zeroconf()
        self._nm_info = None
        self._mm_browser = None

        # Starting managers
        app = self.app
        app.post("/managers/node/start", callback=self.rest_startNM)
        app.post("/managers/node/stop", callback=self.rest_stopNM)
        app.post("/managers/island/start", callback=self.rest_startDIM)
        app.post("/managers/island/stop", callback=self.rest_stopDIM)
        app.post("/managers/master/start", callback=self.rest_startMM)
        app.post("/managers/master/stop", callback=self.rest_stopMM)

        # Querying about managers
        app.get("/managers", callback=self.rest_getMgrs)
        app.get("/managers/master", callback=self.rest_getMMInfo)
        app.get("/managers/island", callback=self.rest_getDIMInfo)
        app.get("/managers/node", callback=self.rest_getNMInfo)

        # Automatically start those that we need
        if master:
            self.startMM()
        if not noNM:
            self.startNM()

    def stop(self, timeout=None):
        """
        Stops this DALiuGE Daemon, terminating all its child processes and its REST
        server.
        """
        self._shutting_down = True
        super(DlgDaemon, self).stop(timeout)
        self.stopNM(timeout)
        self.stopDIM(timeout)
        self.stopMM(timeout)
        self._stop_zeroconf()
        logger.info("DALiuGE Daemon stopped")

    def _stop_zeroconf(self):

        if not self._zeroconf:
            return

        # Stop the MM service browser, the NM registration, and ZC itself
        if self._mm_browser:
            self._mm_browser.cancel()
            self._mm_browser.join()
        self._zeroconf.close()
        logger.info("Zeroconf stopped")

    def _stop_rest_server(self, timeout):
        if self._ioloop:

            self.app.close()

            # Submit a callback to the IOLoop to stop itself and wait until it's
            # done with it
            logger.debug("Stopping the web server")
            ioloop_stopped = threading.Event()

            def stop_ioloop():
                self._ioloop.stop()
                ioloop_stopped.set()

            self._ioloop.add_callback(stop_ioloop)
            if not ioloop_stopped.wait(timeout):
                logger.warning("Timed out while waiting for the server to stop")
            self._server.stop()

            self._started = False

    def _stop_manager(self, name, timeout):
        proc = getattr(self, name)
        if proc:
            utils.terminate_or_kill(proc, timeout)
            pid = proc.pid
            setattr(self, name, None)
            return {"terminated": pid}

    def stopNM(self, timeout=10):
        if self._nm_info:
            utils.deregister_service(self._zeroconf, self._nm_info)
        return self._stop_manager("_nm_proc", timeout)

    def stopDIM(self, timeout=10):
        self._stop_manager("_dim_proc", timeout)

    def stopMM(self, timeout=10):
        self._stop_manager("_mm_proc", timeout)

    # Methods to start and stop the individual managers
    def startNM(self):
        tool = get_tool()
        args = ["--host", "0.0.0.0"]
        args += self._verbosity_as_cmdline()
        logger.info("Starting Node Drop Manager with args: %s" % (" ".join(args)))
        self._nm_proc = tool.start_process("nm", args)
        logger.info("Started Node Drop Manager with PID %d" % (self._nm_proc.pid))

        # Registering the new NodeManager via zeroconf so it gets discovered
        # by the Master Manager
        if self._zeroconf:
            addrs = utils.get_local_ip_addr()
            logger.info("Registering this NM with zeroconf: %s" % addrs)
            self._nm_info = utils.register_service(
                self._zeroconf,
                "NodeManager",
                socket.gethostname(),
                addrs[0][0],
                constants.NODE_DEFAULT_REST_PORT,
            )
        return

    def startDIM(self, nodes):
        tool = get_tool()
        args = ["--host", "0.0.0.0"]
        args += self._verbosity_as_cmdline()
        if nodes:
            args += ["--nodes", ",".join(nodes)]
        logger.info(
            "Starting Data Island Drop Manager with args: %s" % (" ".join(args))
        )
        self._dim_proc = tool.start_process("dim", args)
        logger.info(
            "Started Data Island Drop Manager with PID %d" % (self._dim_proc.pid)
        )
        return

    def startMM(self):
        tool = get_tool()
        args = ["--host", "0.0.0.0"]
        args += self._verbosity_as_cmdline()
        logger.info("Starting Master Drop Manager with args: %s" % (" ".join(args)))
        self._mm_proc = tool.start_process("mm", args)
        logger.info("Started Master Drop Manager with PID %d" % (self._mm_proc.pid))

        # Also subscribe to zeroconf events coming from NodeManagers and feed
        # the Master Manager with the new hosts we find
        if self._zeroconf:
            mm_client = client.MasterManagerClient()
            node_managers = {}

            def nm_callback(zeroconf, service_type, name, state_change):
                info = zeroconf.get_service_info(service_type, name)
                if state_change is zc.ServiceStateChange.Added:
                    server = socket.inet_ntoa(_get_address(info))
                    port = info.port
                    node_managers[name] = (server, port)
                    logger.info(
                        "Found a new Node Manager on %s:%d, will add it to the MM"
                        % (server, port)
                    )
                    mm_client.add_node(server)
                elif state_change is zc.ServiceStateChange.Removed:
                    server, port = node_managers[name]
                    logger.info(
                        "Node Manager on %s:%d disappeared, removing it from the MM"
                        % (server, port)
                    )

                    # Don't bother to remove it if we're shutting down. This way
                    # we avoid hanging in here if the MM is down already but
                    # we are trying to remove our NM who has just disappeared
                    if not self._shutting_down:
                        try:
                            mm_client.remove_node(server)
                        finally:
                            del node_managers[name]

            self._mm_browser = utils.browse_service(
                self._zeroconf, "NodeManager", "tcp", nm_callback
            )
            logger.info("Zeroconf started")
            return

    def _verbosity_as_cmdline(self):
        if self._verbosity > 0:
            return ["-" + "v" * self._verbosity]
        elif self._verbosity < 0:
            return ["-" + "q" * (-self._verbosity)]
        return ()

    # Rest interface
    def _rest_start_manager(self, proc, start_method):
        if proc is not None:
            bottle.abort(409, "The Drop Manager is already running")  # Conflict
        start_method()
        return

    def _rest_stop_manager(self, proc, stop_method):
        if proc is None:
            bottle.abort(409, "The Drop Manager is not running")  # Conflict
        return json.dumps(stop_method())

    def _rest_get_manager_info(self, proc):
        if proc:
            bottle.response.content_type = "application/json"
            logger.info("Sending response: %s" % json.dumps({"pid": proc.pid}))
            return json.dumps({"pid": proc.pid})
        else:
            return json.dumps({"pid": None})

    def rest_getMgrs(self):
        mgrs = {
            "master": self._mm_proc,
            "island": self._dim_proc,
            "node": self._nm_proc,
        }
        if mgrs["master"]:
            mgrs["master"] = self._mm_proc.pid
        if mgrs["island"]:
            mgrs["island"] = self._dim_proc.pid
        if mgrs["node"]:
            mgrs["node"] = self._nm_proc.pid

        logger.info("Sending response: %s" % json.dumps(mgrs))
        return json.dumps(mgrs)

    def rest_startNM(self):
        self._rest_start_manager(self._nm_proc, self.startNM)
        return self.rest_getNMInfo()

    def rest_stopNM(self):
        self._rest_stop_manager(self._nm_proc, self.stopNM)

    def rest_startDIM(self):
        body = bottle.request.json
        if not body or "nodes" not in body:
            # if nothing else is specified we simply add this host
            nodes = {"localhost"}
        else:
            nodes = body["nodes"]
        self._rest_start_manager(
            self._dim_proc, functools.partial(self.startDIM, nodes)
        )
        return self.rest_getDIMInfo()

    def rest_stopDIM(self):
        self._rest_stop_manager(self._dim_proc, self.stopDIM)

    def rest_startMM(self):
        self._rest_start_manager(self._mm_proc, self.startMM)
        return self.rest_getMMInfo()

    def rest_stopMM(self):
        self._rest_stop_manager(self._mm_proc, self.stopMM)

    def rest_getNMInfo(self):
        return self._rest_get_manager_info(self._nm_proc)

    def rest_getDIMInfo(self):
        return self._rest_get_manager_info(self._dim_proc)

    def rest_getMMInfo(self):
        return self._rest_get_manager_info(self._mm_proc)


terminating = False


def run_with_cmdline(parser, args):
    parser.add_option(
        "-m",
        "--master",
        action="store_true",
        dest="master",
        help="Start this DALiuGE daemon as the master daemon",
        default=False,
    )
    parser.add_option(
        "--no-nm",
        action="store_true",
        dest="noNM",
        help="Don't start a NodeDropManager by default",
        default=False,
    )
    parser.add_option(
        "--no-zeroconf",
        action="store_true",
        dest="noZC",
        help="Don't enable zeroconf on this DALiuGE daemon",
        default=False,
    )
    parser.add_option(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        help="Become more verbose. The more flags, the more verbose",
        default=0,
    )
    parser.add_option(
        "-q",
        "--quiet",
        action="count",
        dest="quiet",
        help="Be less verbose. The more flags, the quieter",
        default=0,
    )

    (opts, args) = parser.parse_args(args)

    # -v and -q are exclusive
    if opts.verbose and opts.quiet:
        parser.error("-v and -q cannot be specified together")

    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

    daemon = DlgDaemon(opts.master, opts.noNM, opts.noZC, opts.verbose - opts.quiet)

    # Signal handling, which stops the daemon
    def handle_signal(signalNo, stack_frame):
        global terminating
        if terminating:
            return
        logger.info("Received signal %d, will stop the daemon now" % (signalNo,))
        terminating = True
        daemon.stop(10)

    signal.signal(signal.SIGINT, handle_signal)
    signal.signal(signal.SIGTERM, handle_signal)

    # Go, go, go!
    t = threading.Thread(
        target=daemon.start, args=("0.0.0.0", constants.DAEMON_DEFAULT_REST_PORT)
    )
    t.start()
    signal.pause()
